import dask
from dask.distributed import Client
from prefect import flow, get_run_logger, task

import coiled


@task
def load_data():
    """Load some data"""
    return dask.datasets.timeseries("2000", "2005", partition_freq="2w").persist()


@task
def summarize(df):
    """Compute a summary table"""
    return df.groupby("name").aggregate({"x": "sum", "y": "max"}).compute()


@task
def log_summary(df):
    """Log summary result"""
    logger = get_run_logger()
    logger.info(df)


@flow
def timeseries_flow():
    df = load_data()
    summary_df = summarize(df)
    log_summary(summary_df)


if __name__ == "__main__":
    with coiled.Cluster(name="prefect-timeseries", n_workers=5) as cluster:
        client = Client(cluster)
        timeseries_flow()



from prefect import flow, task
from prefect_dask import DaskTaskRunner

coiled_runner = DaskTaskRunner(
    cluster_class="coiled.Cluster",
    cluster_kwargs={
        "n_workers": 5,
        "software": "prefect2-example",
        "name": "prefect-dask-runner",
    },
)


@task
def inc(x):
    return x + 1


@task
def double(x):
    return x * 2


@task
def add(x, y):
    return x + y


@task
def sum_all(z):
    return sum(z)


@flow(task_runner=coiled_runner)
def inc_double_add_sum(a_range):
    a = inc.map(a_range)
    b = double.map(a_range)
    c = add.map(x=a, y=b)
    return sum_all(c)


if __name__ == "__main__":
    inc_double_add_sum(range(10))



import pytest
from distributed.utils_test import loop  # noqa: F401
from django.test import override_settings

import coiled
from api_tokens.models import ApiToken


@pytest.fixture(
    params=[
        {},
        {"days_to_expire": 2},
        {"label": "my token"},
    ]  # testing out a few different token creation kwargs
)
def created_api_token(sample_user, request):
    with override_settings(LONG_LIVED_TOKENS_ENABLED=True):
        api_token = coiled.create_api_token(**request.param)
    return api_token


@override_settings(LONG_LIVED_TOKENS_ENABLED=True)
def test_can_create_api_token(sample_user, created_api_token):
    assert created_api_token["token"] is not None


@override_settings(LONG_LIVED_TOKENS_ENABLED=True)
def test_default_list_excludes_revoked(sample_user):
    created_api_token = coiled.create_api_token()
    identifier = created_api_token["identifier"]
    coiled.revoke_api_token(identifier=identifier)
    listed_api_tokens = coiled.list_api_tokens()
    assert identifier not in listed_api_tokens


@override_settings(LONG_LIVED_TOKENS_ENABLED=True)
def test_list_create_api_tokens(created_api_token, sample_user):
    """Check that the listed details of an API token match what we created it with.

    Here we'll just check that the tokens *include* the token we expect. It's okay if there are extras.
    We have stricter tests elsewhere (api_tokens/tests.py::test_list_tokens) that confirm the list is
    exactly what we expect, and that users don't see other users' tokens in the list.
    """
    identifier = created_api_token["identifier"]
    listed_api_tokens = coiled.list_api_tokens()
    listed_api_token = listed_api_tokens[identifier]

    assert "token" not in listed_api_token  # the token value shouldn't be here

    # check that the created token and listed token match (except for 'token')
    created_token_without_secret_token = {k: v for k, v in created_api_token.items() if k != "token"}

    assert listed_api_tokens[identifier] == created_token_without_secret_token


@override_settings(LONG_LIVED_TOKENS_ENABLED=True)
def test_revoke_api_token_by_identifier(sample_user):
    created_api_token = coiled.create_api_token()
    identifier = created_api_token["identifier"]
    coiled.revoke_api_token(identifier=identifier)
    listed_api_tokens = coiled.list_api_tokens(include_inactive=True)
    assert listed_api_tokens[identifier]["revoked"]


@override_settings(LONG_LIVED_TOKENS_ENABLED=True)
def test_revoke_api_token_by_label(sample_user):
    token_to_revoke = coiled.create_api_token(label="my token to revoke")
    token_to_keep = coiled.create_api_token(label="my token not to revoke")
    coiled.revoke_api_token(label="my token to revoke")
    listed_api_tokens = coiled.list_api_tokens(include_inactive=True)

    assert listed_api_tokens[token_to_revoke["identifier"]]["revoked"]
    assert not listed_api_tokens[token_to_keep["identifier"]]["revoked"]


@override_settings(LONG_LIVED_TOKENS_ENABLED=True)
def test_revoke_all_api_tokens(sample_user):
    coiled.create_api_token()
    coiled.create_api_token()
    coiled.revoke_all_api_tokens()

    tokens = ApiToken.objects.filter(user_id=sample_user.user.id)
    for token in tokens:
        assert token.revoked



from coiled.cli.core import cli


def test_available_commands():
    assert set(cli.commands) == {
        "cluster",
        "login",
        "setup",
        "env",
        "diagnostics",
        "package-sync",
        "curl",
    }



from __future__ import annotations

import datetime
from collections import Counter
from textwrap import dedent
from types import TracebackType
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Type

import jmespath
import rich
import rich.align
import rich.bar
import rich.box
import rich.console
import rich.layout
import rich.live
import rich.panel
import rich.progress
import rich.table

from ...errors import ClusterCreationError
from ...magic import ResolvedPackageInfo
from ...utils import get_details_url
from ..states import CombinedProcessStateEnum, ProcessStateEnum
from .util import get_instance_types, get_worker_statuses

LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

# This is annoying. I think Ian explained to me that the reason
# we need to manage the size so manually is that if we aren't explicit,
# Jupyter notebooks (but not IPython terminal) messes up the size.
WAITING_CONSOLE_HEIGHT = 26
WAITING_PROGRESS_HEIGHT = 9
DONE_PROGRESS_HEIGHT = 6
DONE_CONSOLE_HEIGHT = WAITING_CONSOLE_HEIGHT - (WAITING_PROGRESS_HEIGHT - DONE_PROGRESS_HEIGHT)
CONSOLE_WIDTH = 100


def level_to_str(int):
    if int >= 100:
        return "Critical"
    elif int >= 50:
        return "Warning"
    elif int >= 0:
        return "Low"


def print_rich_package_table(
    packages_with_notes: List[ResolvedPackageInfo],
    packages_with_errors: List[Tuple[ResolvedPackageInfo, int]],
):
    console = rich.console.Console(width=CONSOLE_WIDTH)

    if packages_with_notes:
        note_table = rich.table.Table(expand=True, box=rich.box.MINIMAL)
        note_table.add_column("Package")
        note_table.add_column("Note", overflow="fold")
        for pkg in packages_with_notes:
            note_table.add_row(pkg["name"], pkg["note"])
        console.print(rich.panel.Panel(note_table, title="[bold green]Package Info[/bold green]"))
    if packages_with_errors:
        error_table = rich.table.Table(expand=True, box=rich.box.MINIMAL)
        error_table.add_column("Package")
        error_table.add_column("Error")
        error_table.add_column("Risk")
        for pkg_info, level in sorted(packages_with_errors, key=lambda p: (p[1], p[0]["name"]), reverse=True):
            error_table.add_row(pkg_info["name"], pkg_info["error"], level_to_str(level))
        console.print(rich.panel.Panel(error_table, title="[bold red]Not Synced with Cluster[/bold red]"))


class RichClusterWidget:
    """A Rich renderable showing cluster status."""

    n_workers: int
    _cluster_details: Mapping[str, Any] | None
    _progress: rich.progress.Progress

    def __init__(
        self,
        n_workers: int = 0,
        transient: bool = False,
        console: rich.console.Console | None = None,
        *,
        server,
        account,
    ):
        """Set up the renderable widget."""
        self.server = server
        self.account = account
        self._cluster_details = None
        self._final_update = None
        self._loader_frames = ["...", " ..", "  .", "   ", ".  ", ".. "]
        self._frame = 0
        self.last_updated_utc = datetime.datetime.now(datetime.timezone.utc)
        if console:
            self.console = console
        else:
            self.console = rich.console.Console()
            self.console.size = (CONSOLE_WIDTH, WAITING_CONSOLE_HEIGHT)

        self._progress = progress = rich.progress.Progress(
            "{task.description}",
            rich.progress.BarColumn(complete_style="progress.remaining"),
            rich.progress.TextColumn("[progress.percentage]{task.completed}"),
        )
        self._error_progress = error_progress = rich.progress.Progress(
            "{task.description}",
            rich.progress.BarColumn(complete_style="red", finished_style="red"),
            rich.progress.TextColumn("[progress.percentage]{task.completed}"),
        )
        self._provision_task = progress.add_task("Provisioning", total=n_workers)
        self._boot_task = progress.add_task("Booting Instance", total=n_workers)

        self._downloading_extracting_task = progress.add_task("Downloading environment", total=n_workers)
        self._ready_task = progress.add_task("Ready", total=n_workers)
        self._stopping_task = progress.add_task("Stopping", total=n_workers)
        self._stopped_task = progress.add_task("Stopped", total=n_workers)

        # In order to make the bar red for errors, the error progress bar is a different instance
        # of rich.progress.Progress.
        # But we still want this bar aligned with the others, since rich doesn't know these two groups
        # are related, it doesn't automatically know how much right-padding to add
        # after the word "Error", so we calculate that ourselves here.
        max_len = max(len(t.description) for t in progress.tasks)
        self._error_task = error_progress.add_task(
            "Error" + " " * (max_len - len("Error")), total=n_workers, visible=False
        )

        # Calls _get_renderable, so the progress bar must be set up first.
        self.live = rich.live.Live(
            transient=transient,
            console=self.console,
            get_renderable=self._get_renderable,
        )

    def start(self):
        """Start the live instance."""
        self.live.start(refresh=False)

    def stop(self):
        """Stop the live instance."""
        self.live.stop()

    def __enter__(self) -> RichClusterWidget:
        """Enter a live-updating context.

        Example
        -------
        with RichClusterWidget(n_workers) as w:
            # do stuff
        """
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the live-updating context and reset the display ID.

        Keep the widget around for user inspection if there was a cluster creation
        error, otherwise remove it.
        """
        if exc_type == ClusterCreationError and self.live.transient:
            self.live.transient = False
        self.stop()

    def update(
        self,
        cluster_details: Optional[Mapping[str, Any]],
        logs,
        *args,
        final_update=None,
        **kwargs,
    ) -> None:
        """Update cluster data.

        Note: this does not trigger any refreshing, that is handled by the Live
        instance, which does it periodically.
        """
        self._cluster_details = cluster_details

        # track when the widget last received new data about the cluster
        # (which is less often than the widget renders)
        self.last_updated_utc = datetime.datetime.now(datetime.timezone.utc)

        if final_update:
            self._final_update = final_update

        # We explicitly refresh to make sure updated info is shown.
        # (Bad timing can lead auto-refresh to not show the update before we stop.)
        self.live.refresh()

    def _ipython_display_(self) -> None:
        """Render in a notebook context.

        Note: this is *not* called in an IPython terminal context. Instead,
        _repr_mimebundle_ is used in the IPython terminal.
        """
        self.console.print(self._get_renderable(), new_line_start=True)

    def __rich_console__(
        self, console: rich.console.Console, options: rich.console.ConsoleOptions
    ) -> rich.console.RenderResult:
        """Implement the Rich console interface.

        In particular, this is used in ``_repr_mimebundle_`` to display to an IPython
        terminal.
        """
        yield self._get_renderable()

    def _repr_mimebundle_(self, include: Iterable[str], exclude: Iterable[str], **kwargs) -> dict[str, str]:
        """Display the widget in an IPython console context.

        This is adapted from the code in `rich.jupyter.JupyterMixin`. We can't
        use that mixin because it doesn't allow you to specify your own console
        (instead using the global one). We want our own console because we
        manually set the size to not take up the full terminal.
        """
        console = self.console
        segments = list(console.render(self, console.options))  # type: ignore
        text = console._render_buffer(segments)  # Unfortunate private member access...
        data = {"text/plain": text}
        if include:
            data = {k: v for (k, v) in data.items() if k in include}
        if exclude:
            data = {k: v for (k, v) in data.items() if k not in exclude}
        return data

    def _get_renderable(self) -> rich.console.RenderableType:
        """Construct the rendered layout."""
        progress = self._progress
        error_progress = self._error_progress
        current_loader_frame = self._loader_frames[self._frame]
        if self._cluster_details:
            desired_workers = self._cluster_details["desired_workers"]
            n_workers = len(jmespath.search("workers[*]", self._cluster_details) or [])
            overall_cluster_status = self._cluster_details["current_state"]["state"]
            if self._cluster_details["current_state"]["state"] != "ready":
                overall_cluster_status += current_loader_frame

            scheduler_status = self._cluster_details["scheduler"]["current_state"]["state"]
            scheduler_ready = ProcessStateEnum(scheduler_status) == ProcessStateEnum.started
            if not scheduler_ready:
                scheduler_status += current_loader_frame

            dashboard_address = (
                jmespath.search("scheduler.dashboard_address", self._cluster_details) or None
                if scheduler_ready
                else None
            )

            scheduler_instance_type, worker_instance_types = get_instance_types(self._cluster_details)
            region = self._cluster_details["cluster_options"]["region_name"]
            zone = self._cluster_details["cluster_options"]["zone_name"]
            zone_desc = f" ({zone})" if zone else ""
            cluster_name = self._cluster_details["name"]

            statuses = get_worker_statuses(self._cluster_details)
            progress.update(
                self._provision_task,
                total=n_workers,
                completed=statuses[CombinedProcessStateEnum.instance_queued]
                + statuses[CombinedProcessStateEnum.instance_starting],
            )
            progress.update(
                self._downloading_extracting_task,
                total=n_workers,
                completed=statuses[CombinedProcessStateEnum.downloading],
            )
            progress.update(
                self._boot_task,
                total=n_workers,
                completed=statuses[CombinedProcessStateEnum.instance_running],
            )
            progress.update(
                self._ready_task,
                total=n_workers,
                completed=statuses[CombinedProcessStateEnum.ready],
            )
            progress.update(
                self._stopping_task,
                total=n_workers,
                completed=statuses[CombinedProcessStateEnum.stopping],
            )
            progress.update(
                self._stopped_task,
                total=n_workers,
                completed=statuses[CombinedProcessStateEnum.stopped],
            )
            error_progress.update(
                self._error_task,
                visible=statuses[CombinedProcessStateEnum.error] > 0,
                total=n_workers,
                completed=statuses[CombinedProcessStateEnum.error],
            )
        else:
            overall_cluster_status = current_loader_frame
            scheduler_status = current_loader_frame
            dashboard_address = None
            scheduler_instance_type = current_loader_frame
            worker_instance_types = Counter()
            region = current_loader_frame
            zone_desc = ""
            cluster_name = ""
            desired_workers = ""

        dashboard_label = (
            f"[link={dashboard_address}]{dashboard_address}[/link]" if dashboard_address else current_loader_frame
        )
        if any(k for k, v in worker_instance_types.items()):
            worker_instance_types_label = ", ".join(
                f"{k or 'Unknown'} ({v:,})" for k, v in worker_instance_types.items()
            )
        else:
            worker_instance_types_label = current_loader_frame
        config = dedent(
            f"""
            [bold green]Region:[/bold green] {region}{zone_desc}

            [bold green]Scheduler Instance Type:[/bold green] {scheduler_instance_type or current_loader_frame}

            [bold green]Worker Instance Type(s):[/bold green] {worker_instance_types_label or current_loader_frame}

            [bold green]Workers Requested:[/bold green] {desired_workers}"""
        )

        status = dedent(
            f"""
            [bold green]Cluster Name:[/bold green] {cluster_name}

            [bold green]Cluster Status:[/bold green] {overall_cluster_status}

            [bold green]Scheduler Status:[/bold green] {scheduler_status}

            [bold green]Dashboard Address:[/bold green] {dashboard_label}"""
        )

        """Define the layout."""
        layout = rich.layout.Layout(name="root")
        self._frame = (self._frame + 1) % len(self._loader_frames)

        if self._final_update is None:
            progress_height = WAITING_PROGRESS_HEIGHT
            console_size = (CONSOLE_WIDTH, WAITING_CONSOLE_HEIGHT)
        else:
            progress_height = DONE_PROGRESS_HEIGHT
            console_size = (CONSOLE_WIDTH, DONE_CONSOLE_HEIGHT)

        self.console.size = console_size

        if self._cluster_details is None:
            link = ""
        else:
            link = get_details_url(self.server, self.account, self._cluster_details["id"])
            assert link is not None  # typechecker, go away please
        layout.split(
            rich.layout.Layout(
                rich.panel.Panel(
                    rich.align.Align.center(f"[link]{link}[/link]"),
                    title="[bold green frame]Coiled Cluster",
                ),
                name="header",
                size=3,
            ),
            rich.layout.Layout(name="body", size=12),
            rich.layout.Layout(name="progress", size=progress_height),
        )
        layout["body"].split_row(
            rich.layout.Layout(name="overview"),
            rich.layout.Layout(name="configuration"),
        )

        time = self.last_updated_utc.astimezone(LOCAL_TIMEZONE).strftime("%Y/%m/%d %H:%M:%S %Z")
        if self._final_update is None:
            layout["progress"].update(
                rich.panel.Panel(
                    rich.align.Align.center(
                        rich.console.Group(progress.get_renderable(), error_progress.get_renderable()),
                        vertical="middle",
                    ),
                    title=f"Dask Worker States ({time})",
                )
            )
        else:
            layout["progress"].update(
                rich.panel.Panel(
                    rich.align.Align.center(self._final_update, vertical="middle"),
                    title=f"({time})",
                )
            )
        layout["body"]["overview"].update(rich.panel.Panel(status, title="Overview"))
        layout["body"]["configuration"].update(rich.panel.Panel(config, title="Configuration"))
        return layout



import os
import shutil
import subprocess
import tempfile
from typing import Optional

import click
from rich import print

import coiled

from ..utils import CONTEXT_SETTINGS


def open_ssh(address: str, key: str, jump_to_address: Optional[str] = None):
    """Open an SSH session, relies on `ssh` and `ssh-add` (agent)."""

    if not shutil.which("ssh"):
        print("Unable to find `ssh`, you may need to install OpenSSH or add it to your paths.")
        return

    if not shutil.which("ssh-add"):
        print("Unable to find `ssh-add`, you may need to install OpenSSH or add it to your paths.")
        return

    with tempfile.TemporaryDirectory() as keydir:
        key_path = os.path.join(keydir, f"scheduler-key-{address}.pem")

        with open(key_path, "w") as f:
            f.write(key)

        # ssh needs file permissions to be set
        os.chmod(key_path, mode=0o600)

        # briefly add key to agent, this allows us to jump to worker with agent forwarding
        p = subprocess.run(["ssh-add", "-t", "5", key_path], capture_output=True)

        if p.returncode and p.stderr:
            print("An error occurred calling `ssh-add`. You may need to enable the ssh agent.")
            print(p.stderr)
            return

    if jump_to_address:
        ssh_target = f"-J ubuntu@{address} ubuntu@{jump_to_address}"
        ssh_target_label = f"worker at {jump_to_address}"
    else:
        ssh_target = f"ubuntu@{address}"
        ssh_target_label = f"scheduler at {address}"

    ssh_command = f"ssh {ssh_target} -o StrictHostKeyChecking=no -o ForwardAgent=yes"

    # print(ssh_command)
    print(f"===Starting SSH session to {ssh_target_label}===")
    subprocess.run(ssh_command, shell=True)
    print("===SSH session closed===")


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("cluster")
@click.option(
    "--private",
    default=False,
    is_flag=True,
    help="Use private IP address of scheduler (default is public IP address)",
)
@click.option(
    "--worker",
    default=None,
    help="Connect to worker with specified name or private IP address (default is to connect to scheduler)",
)
def ssh(cluster: str, private: bool, worker: Optional[str]):
    with coiled.Cloud() as cloud:
        if cluster.isnumeric():
            cluster_id = int(cluster)
        else:
            try:
                cluster_id = cloud.get_cluster_by_name(name=cluster)
            except coiled.errors.DoesNotExist:
                cluster_id = None

        if not cluster_id:
            print(f"Unable to find cluster `{cluster}`")
            return

        ssh_info = cloud.get_ssh_key(cluster_id=cluster_id, worker=worker)

        # print(ssh_info)

        if ssh_info["scheduler_state"] not in ("starting", "started"):
            print(
                f"Scheduler state is {ssh_info['scheduler_state']}. "
                "You can only connect when scheduler is 'starting' or 'started'."
            )
            return

        scheduler_address = ssh_info["scheduler_private_address"] if private else ssh_info["scheduler_public_address"]

        if not scheduler_address:
            print("Unable to retrieve scheduler address")
            return

        if not ssh_info["private_key"]:
            print("Unable to retrieve SSH key")
            return

        open_ssh(
            address=scheduler_address,
            key=ssh_info["private_key"],
            jump_to_address=ssh_info["worker_address"],
        )



import re
import time
from datetime import datetime, timedelta, timezone
from typing import List, Optional

import click
from rich.console import Console

import coiled

from ..utils import CONTEXT_SETTINGS

COLORS = [
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
]


@click.command(
    context_settings=CONTEXT_SETTINGS,
)
@click.option(
    "--account",
    default=None,
    help="Coiled account (uses default account if not specified)",
)
@click.option(
    "--cluster",
    default=None,
    help="Cluster for which to show logs, default is most recent",
)
@click.option(
    "--no-scheduler",
    default=False,
    is_flag=True,
    help="Don't include scheduler logs",
)
@click.option(
    "--workers",
    default="all",
    help=(
        "All worker logs included by default, specify 'none' or "
        "comma-delimited list of names, states, or internal IP addresses"
    ),
)
@click.option(
    "--label",
    default="private_ip_address",
    type=click.Choice(
        ["private_ip_address", "name", "id", "public_ip_address", "none"],
        case_sensitive=False,
    ),
)
@click.option(
    "--system",
    default=False,
    is_flag=True,
    help="Just show system logs",
)
@click.option(
    "--combined",
    default=False,
    is_flag=True,
    help="Show combined system and dask logs",
)
@click.option(
    "--tail",
    default=False,
    is_flag=True,
    help="Keep tailing logs",
)
@click.option(
    "--since",
    default=None,
    help="By default will show logs from start of cluster (or 30s ago if tailing)",
)
@click.option(
    "--until",
    default=None,
    help="Show logs up to and including this time, by default will go through present time.",
)
@click.option(
    "--filter",
    default=None,
    help="Filter log messages",
)
@click.option(
    "--color/--no-color",
    default=True,
    is_flag=True,
    help="Use for color in logs",
)
@click.option(
    "--show-all-timestamps",
    default=False,
    is_flag=True,
    help="Prepend datetime to all log messages",
)
@click.option(
    "--interval",
    default=3,
    help="Tail polling interval",
)
def better_logs(
    account: Optional[str],
    cluster: Optional[str],
    no_scheduler: bool,
    workers: str,
    label: str,
    system: bool,
    combined: bool,
    tail: bool,
    interval: int,
    since: Optional[str],
    until: Optional[str],
    color: bool,
    filter: Optional[str],
    show_all_timestamps: bool,
):
    console = Console(force_terminal=color)

    dask = not system or combined
    system = system or combined
    label = label.lower()

    if tail and until:
        raise click.ClickException("You can't use --until when tailing logs.")

    with coiled.Cloud(account=account) as cloud:
        if cluster and cluster.isnumeric():
            cluster_id = int(cluster)
        elif cluster:
            # get cluster by name
            try:
                clusters = cloud.get_clusters_by_name(name=cluster)
                if clusters:
                    recent_cluster = clusters[-1]
                else:
                    raise click.ClickException(f"Unable to find cluster with name '{cluster}'")

                if tail and recent_cluster["current_state"]["state"] in (
                    "stopped",
                    "error",
                ):
                    tail = False
                    console.print(
                        f"[red]Cluster state is {recent_cluster['current_state']['state']} so not tailing.[/red]"
                    )

                cluster_id = recent_cluster["id"]

            except coiled.errors.DoesNotExist:
                cluster_id = None
        else:
            # default to most recent cluster
            clusters = coiled.list_clusters(max_pages=1)
            if not clusters:
                raise ValueError("Unable to find any clusters for your account")
            match = max(clusters, key=lambda c: c["id"])
            cluster_id = match["id"]

        if not cluster_id:
            raise click.ClickException(f"Unable to find cluster `{cluster}`")

        cluster_info = cloud.cluster_details(cluster_id)

    # instance ID's for which to show logs, key maps to label to use
    instances = {}
    if not no_scheduler and cluster_info.get("scheduler", {}).get("instance"):
        scheduler_id = cluster_info["scheduler"]["instance"]["id"]
        instances[scheduler_id] = {
            "label": "scheduler" if label != "none" else "",
            "color": COLORS[-1],
        }

    def worker_label(worker: dict):
        if label == "none":
            return ""
        return (
            worker.get("name", str(worker["instance"]["id"]))
            if label == "name"
            else worker["instance"].get(label, str(worker["instance"]["id"]))
        )

    # TODO when tailing "all" workers, this won't include workers that appear after we start
    #  (addressing this is future enhancement)

    if workers:
        worker_attrs_to_match = workers.split(",")

        def filter_worker(worker):
            if worker.get("name") and worker["name"] in worker_attrs_to_match:
                # match on name
                return True
            elif (
                worker.get("instance", {}).get("private_ip_address")
                and worker["instance"]["private_ip_address"] in worker_attrs_to_match
            ):
                # match on private IP
                return True
            elif (
                worker.get("current_state", {}).get("state")
                and worker["current_state"]["state"] in worker_attrs_to_match
            ):
                # match on state
                return True

            return False

        instances.update(
            {
                worker["instance"]["id"]: dict(label=worker_label(worker), color=COLORS[idx % len(COLORS)])
                for idx, worker in enumerate(cluster_info["workers"])
                if worker.get("instance") and (workers == "all" or filter_worker(worker))
            }
        )

    show_all_instances = workers == "all" and not no_scheduler

    console.print(f"=== Logs for {cluster_info['name']} ({cluster_id}) ===\n")

    if not instances:
        # exit here if there are no instances, otherwise passing empty list to
        # coiled.better_cluster_logs will result in pulling logs for all instances
        console.print("no instances match specified filters")
        return

    from_timestamp = ts_ms_from_string(since)
    until_timestamp = ts_ms_from_string(until)
    last_events = set()

    if tail and not from_timestamp:
        # for tail, start with logs from 30s ago if start isn't specified
        current_ms = int(time.time_ns() // 1e6)
        from_timestamp = current_ms - (30 * 1000)

    while True:
        events = coiled.better_cluster_logs(
            cluster_id=cluster_id,
            # function returns all instances if none specified, we'll use that
            # in order to not exclude instances that show up while tailing
            instance_ids=None if show_all_instances else list(instances.keys()),
            dask=dask,
            system=system,
            since_ms=from_timestamp,
            until_ms=until_timestamp,
            filter=filter,
        )

        if last_events:
            events = [e for e in events if e["timestamp"] != from_timestamp or event_dedupe_key(e) not in last_events]

        if events:
            print_events(console, events, instances, show_all_timestamps=show_all_timestamps)

            from_timestamp = events[-1]["timestamp"]
            last_events = {event_dedupe_key(e) for e in events if e["timestamp"] == from_timestamp}

        if tail:
            # TODO stop tailing once cluster is stopped/errored (future MR)
            time.sleep(interval)
        else:
            break


def match_cluster(clusters: List[dict], cluster: str) -> dict:
    if cluster.isnumeric():
        matches = [c for c in clusters if c["id"] == int(cluster)]
        if len(matches) == 1:
            return matches[0]
        elif len(matches) > 1:
            raise ValueError(f"Multiple clusters match '{cluster}'")

    # try to match on cluster name
    matches = sorted([c for c in clusters if c["name"] == cluster], key=lambda c: c["id"])
    if matches:
        return matches[-1]

    raise ValueError(f"No clusters match '{cluster}'")


def event_dedupe_key(event):
    return f'{event["timestamp"]}#{event["instance_id"]}#{event["message"]}'


def print_events(console, events, instances: dict, pretty=True, show_all_timestamps=False):
    for e in events:
        console.print(format_log_event(e, instances, pretty=pretty, show_all_timestamps=show_all_timestamps))


def format_log_event(event: dict, instances: dict, pretty: bool, show_all_timestamps: bool) -> str:
    message = event["message"]
    if event["instance_id"] in instances:
        label = instances[event["instance_id"]]["label"]
        color = instances[event["instance_id"]]["color"]
    else:
        # we might not know about instance if it showed up while we're tailing all worker logs
        label = event["instance_id"]
        color = COLORS[0]

    time_string = ""

    if event.get("timestamp"):
        time_format = "%Y-%m-%d %H:%M:%S.%f"
        t = datetime.utcfromtimestamp(event["timestamp"] / 1000)

        if show_all_timestamps or not message_has_timestamp(message, t):
            time_string = f"{t.strftime(time_format)} "

    # indent multiline log messages
    if "\n" in message:
        message = message.replace("\n", "\n  ")

    if label:
        formatted_label = f"[{color}]({label})[/{color}] \t" if pretty else f"({label}) \t"
    else:
        formatted_label = ""

    return f"{formatted_label}{time_string}{message}"


def message_has_timestamp(message: str, t: datetime):
    # naively check if timestamp already present in message by looking for year
    # if it's not in log message, then prepend
    if str(t.year) in message:
        return True


def ts_ms_from_string(timestring: Optional[str]) -> Optional[int]:
    # input can be
    #   int: timestamp (ms)
    #   string in ISO 8601 format
    #   string representing delta (e.g, "1h15m")

    if not timestring:
        return None

    if timestring.isnumeric():
        return int(timestring)

    delta_regex = re.compile(r"^((?P<days>\d+)d)?((?P<hours>\d+)h)?((?P<minutes>\d+)m)?((?P<seconds>\d+)s)?$")
    match = delta_regex.match(timestring)
    if match:
        match_parts = {key: int(val) for key, val in match.groupdict().items() if val}

        delta = timedelta(**match_parts)
        if delta:
            t = datetime.now(tz=timezone.utc) - delta
            return int(t.timestamp() * 1000)

    try:
        t = datetime.fromisoformat(timestring)
        # interpret as UTC if not specified (rather than local time)
        if t.tzinfo is None:
            t = t.replace(tzinfo=timezone.utc)
    except ValueError:
        pass
    else:
        return int(t.timestamp() * 1000)

    raise ValueError(
        f"Unable to convert '{timestring}' into a timestamp, you can use number (timestamp in ms), "
        f"ISO format, or delta such as 5m."
    )



import asyncio
import json
import os
import platform
import sys
import typing
from logging import getLogger
from pathlib import Path
from typing import Dict, List, Optional, Union
from urllib.parse import urlparse

from importlib_metadata import Distribution
from rich.progress import Progress
from typing_extensions import Literal

from coiled.types import CondaPackage, CondaPlaceHolder, PackageInfo
from coiled.utils import parse_file_uri

logger = getLogger("coiled.package_sync")
subdir_datas = {}
PYTHON_VERSION = platform.python_version_tuple()


def consistent_name(name: str):
    # gives the best shot at names being consistent between pypi/conda
    return name.lower().replace("-", "_")


async def scan_conda(prefix: Path, progress: Optional[Progress] = None) -> typing.Dict[str, PackageInfo]:
    conda_meta = prefix / "conda-meta"
    if conda_meta.exists() and conda_meta.is_dir():
        conda_packages = [
            CondaPackage(json.load(metafile.open("r")), prefix=prefix)
            for metafile in conda_meta.iterdir()
            if metafile.suffix == ".json"
        ]
        packages: List[PackageInfo] = []
        if progress:
            for task in progress.track(
                asyncio.as_completed([handle_conda_package(pkg) for pkg in conda_packages]),
                description=f"Scanning {len(conda_packages)} conda packages",
                total=len(conda_packages),
            ):
                packages.append(await task)
        else:
            packages = await asyncio.gather(*[handle_conda_package(pkg) for pkg in conda_packages])
        return {consistent_name(pkg["name"]): pkg for pkg in packages}
    else:
        return {}


async def handle_conda_package(pkg: CondaPackage) -> PackageInfo:
    # Are there conda packages that install multiple python packages?
    metadata_location = next((Path(fp).parent for fp in pkg.files if fp.endswith("METADATA")), None)
    if metadata_location:
        dist = Distribution.at(pkg.prefix / metadata_location)
        name = dist.metadata["Name"] or pkg.name
    else:
        name = pkg.name
    return {
        "channel": pkg.channel,
        "path": None,
        "channel_url": pkg.channel_url,
        "source": "conda",
        "conda_name": pkg.name,
        "subdir": pkg.subdir,
        "name": name,
        "version": pkg.version,
        "wheel_target": None,
    }


async def handle_dist(dist: Distribution, locations: List[Path]) -> Optional[Union[PackageInfo, CondaPlaceHolder]]:
    installer = dist.read_text("INSTALLER") or ""
    installer = installer.rstrip()
    # dist._path can sometimes be a zipp.Path or something else
    dist_path = Path(str(dist._path))  # type: ignore
    if installer == "conda":
        return CondaPlaceHolder(name=dist.name)
    elif dist_path.parent.suffix == ".egg":
        # egg files can be a directory OR a zip file
        # the zipp implementation of Path always uses
        # linux style seperators so we strip them too
        return {
            "name": dist.name,
            "path": dist_path.parent,
            "source": "pip",
            "channel": None,
            "subdir": None,
            "channel_url": None,
            "conda_name": None,
            "version": dist.version,
            "wheel_target": str(dist_path.parent).rstrip(os.sep + "/"),
        }
    else:
        direct_url_metadata = dist.read_text("direct_url.json")
        if direct_url_metadata:
            url_metadata = json.loads(direct_url_metadata)
            url = url_metadata.get("url")
            if not url:
                # no url in this file
                # invalid PEP-610 so don't do anything
                pass
            elif url_metadata.get("vcs_info"):
                # PEP-610 Source is VCS
                vcs_info = url_metadata.get("vcs_info")
                vcs: Literal["git", "hg", "bzr", "svn"] = vcs_info["vcs"]
                commit = vcs_info["commit_id"]
                url = url_metadata["url"]
                pip_url = f"{vcs}+{url}@{commit}"
                return {
                    "name": dist.name,
                    "path": dist_path,
                    "source": "pip",
                    "channel": None,
                    "channel_url": None,
                    "subdir": None,
                    "conda_name": None,
                    "version": dist.version,
                    "wheel_target": pip_url,
                }
            elif str((Path("pypoetry") / "artifacts")) in url_metadata["url"]:
                # if the install source is actually the pre 1.2 poetry cache location
                # they this is actually just normal a pypi
                # and we can ignore direct_url.json
                pass
            elif url_metadata.get("archive_info") is not None:
                # PEP-610 - Source is an archive/wheel, somewhere!
                p = urlparse(url)
                if p.scheme == "file":
                    url = str(parse_file_uri(url))
                return {
                    "name": dist.name,
                    "path": dist_path,
                    "source": "pip",
                    "channel": None,
                    "channel_url": None,
                    "subdir": None,
                    "conda_name": None,
                    "version": dist.version,
                    "wheel_target": url,
                }
            elif url_metadata.get("dir_info") is not None:
                # PEP-610 - Source is a local directory
                path = parse_file_uri(url)
                return {
                    "name": dist.name,
                    "path": path,
                    "source": "pip",
                    "channel": None,
                    "channel_url": None,
                    "subdir": None,
                    "conda_name": None,
                    "version": dist.version,
                    "wheel_target": str(path),
                }
        egg_links = []
        for location in locations:
            if os.name == "nt":
                egg_link_pth = location / Path(dist.name).with_suffix(".egg-link")
                if egg_link_pth.is_file():
                    egg_links.append(location / Path(dist.name).with_suffix(".egg-link"))
            else:
                egg_link_pth = location / Path(dist.name).with_suffix(".egg-link")
                if egg_link_pth.is_file():
                    egg_links.append(location / Path(dist.name).with_suffix(".egg-link"))
        if egg_links:
            return {
                "name": dist.name,
                "path": dist_path.parent,
                "source": "pip",
                "channel": None,
                "channel_url": None,
                "subdir": None,
                "conda_name": None,
                "version": dist.version,
                "wheel_target": str(dist_path.parent),
            }
        return {
            "name": dist.name,
            "path": dist_path,
            "source": "pip",
            "channel": None,
            "channel_url": None,
            "subdir": None,
            "conda_name": None,
            "version": dist.version,
            "wheel_target": None,
        }


async def scan_pip(
    locations: List[Path], progress: Optional[Progress] = None
) -> typing.Dict[str, Union[PackageInfo, CondaPlaceHolder]]:
    # distributions returns ALL distributions
    # even ones that are not active
    # this is a trick so we only get the distribution
    # that is last in stack
    locations = [location for location in locations if location.exists() and location.is_dir()]
    paths: List[str] = [str(location) for location in locations]
    for location in locations:
        for fp in location.iterdir():
            if fp.suffix in [".pth", ".egg-link"]:
                for line in fp.read_text().split("\n"):
                    if line.startswith("#"):
                        continue
                    elif line.startswith(("import", "import\t")):
                        continue
                    elif line.rstrip() == ".":
                        continue
                    else:
                        p = location / Path(line.rstrip())
                        full_path = str(p.resolve())
                        if p.exists() and full_path not in paths:
                            paths.append(full_path)
    active_dists: Dict[str, Distribution] = {dist.name: dist for dist in Distribution.discover(path=list(paths))}
    dists = active_dists.values()
    packages = []
    if progress:
        for task in progress.track(
            asyncio.as_completed([handle_dist(dist, locations) for dist in dists]),
            total=len(dists),
            description=f"Scanning {len(dists)} python packages",
        ):
            packages.append(await task)
    else:
        packages = await asyncio.gather(*(handle_dist(dist, locations) for dist in dists))
    return {consistent_name(pkg["name"]): pkg for pkg in packages if pkg}


async def scan_prefix(
    prefix: Optional[Path] = None, progress: Optional[Progress] = None, locations: Optional[List[Path]] = None
) -> typing.List[PackageInfo]:
    # TODO: private conda channels
    # TODO: detect pre-releases and only set --pre flag for those packages (for conda)

    if not prefix:
        prefix = Path(sys.prefix)
    conda_env_future = asyncio.create_task(scan_conda(prefix=prefix, progress=progress))
    # only pass locations to support testing, otherwise we should be using sys.path
    pip_env_future = asyncio.create_task(
        scan_pip(locations=locations or [Path(p) for p in sys.path], progress=progress)
    )
    conda_env = await conda_env_future
    pip_env = await pip_env_future
    filterd_conda = {}
    # the pip list is the "truth" of what is imported for python deps
    for name, package in conda_env.items():
        # if a package exists in the pip list but is not a conda place holder
        # then the conda package wont be imported and should be discarded
        if pip_env.get(name):
            if isinstance(pip_env[name], CondaPlaceHolder):
                filterd_conda[name] = package
            elif pip_env[name]["version"] == package["version"]:
                pip_env.pop(name, None)
                filterd_conda[name] = package
        else:
            # a non python package and safe to include
            filterd_conda[name] = package
    # remove conda placeholders
    pip_env = {pkg_name: pkg for pkg_name, pkg in pip_env.items() if not isinstance(pkg, CondaPlaceHolder)}
    return sorted(
        list(pip_env.values()) + list(filterd_conda.values()),
        key=lambda pkg: pkg["name"],
    )



import asyncio
from pathlib import Path

from coiled.scan import scan_prefix


def test_scan():
    prefix = Path(__file__).parent / "dummy-env"
    pypath = [
        prefix / "lib" / "python3.9" / "site-packages",
        prefix / "lib" / "python3.9",
    ]
    env = asyncio.run(scan_prefix(prefix=prefix, locations=pypath))
    assert env == [
        {
            "channel": "conda-forge",
            "channel_url": "https://conda.anaconda.org/conda-forge",
            "conda_name": "condapythonpackage",
            "name": "conda_python_package_python_name",
            "path": None,
            "source": "conda",
            "subdir": "noarch",
            "version": "0.1.0",
            "wheel_target": None,
        },
        {
            "channel": "conda-forge",
            "channel_url": "https://conda.anaconda.org/conda-forge",
            "conda_name": "condabinpackage",
            "name": "condabinpackage",
            "path": None,
            "source": "conda",
            "subdir": "osx-arm64",
            "version": "0.0.2",
            "wheel_target": None,
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "dist_info_archive_package",
            "source": "pip",
            "path": prefix / "lib" / "python3.9" / "site-packages" / "dist_info_archive_info_package.dist-info",
            "subdir": None,
            "version": "0.0.10",
            "wheel_target": "https://github.com/pypa/pip/archive/1.3.1.zip",
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "dist_info_dir_info_package",
            "path": Path("/home/user/project/urlencoded@path"),
            "source": "pip",
            "subdir": None,
            "version": "0.0.10",
            "wheel_target": "/home/user/project/urlencoded@path",
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "dist_info_package",
            "path": prefix / "lib" / "python3.9" / "site-packages" / "dist_info_package.dist-info",
            "source": "pip",
            "subdir": None,
            "version": "0.0.11",
            "wheel_target": None,
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "dist_info_vcs_info_package",
            "path": prefix / "lib" / "python3.9" / "site-packages" / "dist_info_vcs_info_package.dist-info",
            "source": "pip",
            "subdir": None,
            "version": "0.0.10",
            "wheel_target": "git+https://github.com/dask/distributed.git@c2dfea237ffe802883b85617f20f7a7ad7b16080",
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "egg_info_package",
            "path": prefix / "lib" / "python3.9" / "site-packages" / "egg_info_package.egg-info",
            "source": "pip",
            "subdir": None,
            "version": "0.0.15",
            "wheel_target": None,
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "egg_link_package",
            "source": "pip",
            "subdir": None,
            "version": "0.0.5",
            "wheel_target": str(prefix / "src" / "egg_link_package"),
            "path": prefix / "src" / "egg_link_package",
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "egg_package",
            "path": prefix / "lib" / "python3.9" / "site-packages" / "egg_package.egg",
            "source": "pip",
            "subdir": None,
            "version": "0.0.20",
            "wheel_target": str(prefix / "lib" / "python3.9" / "site-packages" / "egg_package.egg"),
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "poetry_package",
            "path": Path("/some/where/over/the/rainbow"),
            "source": "pip",
            "subdir": None,
            "version": "0.0.11",
            "wheel_target": "/some/where/over/the/rainbow",
        },
        {
            "channel": None,
            "channel_url": None,
            "conda_name": None,
            "name": "pth_package",
            "path": Path("/some/where/over/the/rainbow"),
            "source": "pip",
            "subdir": None,
            "version": "0.0.25",
            "wheel_target": "/some/where/over/the/rainbow",
        },
    ]



class ServerError(Exception):
    pass


class DoesNotExist(Exception):
    pass


class ClusterCreationError(Exception):
    def __init__(self, message, cluster_id=None):
        self.cluster_id = cluster_id
        super().__init__(message + (f" (cluster_id: {cluster_id})" if self.cluster_id else ""))



import click

from ..utils import login_if_required
from .utils import CONTEXT_SETTINGS


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-s", "--server", help="Coiled server to use")
@click.option("-t", "--token", help="Coiled user token")
@click.option("-a", "--account", help="Coiled account")
@click.option(
    "--retry/--no-retry",
    default=True,
    help="Whether or not to automatically ask for a new token if an invalid token is entered",
)
def login(server, token, account, retry):
    """Configure your Coiled account credentials"""
    login_if_required(server=server, token=token, account=account, save=True, use_config=False, retry=retry)



from typing import Optional


class CoiledException(Exception):
    """Custom exception to be used as a base exception.

    This exception needs to include a `code` argument which is
    a constant variable that will give us more information as to
    where the exception happened.

    """

    code = "COILED_EXCEPTION"

    def __init__(self, message: str, **kwargs):
        super().__init__(message)
        self.message = message
        self.extras = kwargs

    def __repr__(self) -> str:
        return f"CoiledException({self.code}: {self.message})"

    def as_json(self) -> dict:
        return {"code": self.code, "message": self.message, **self.extras}


class UnsupportedBackendError(CoiledException):
    code = "UNSUPPORTED_BACKEND"


class ParameterMissingError(CoiledException):
    code = "PARAMETER_MISSING_ERROR"


class GCPCredentialsParameterError(ParameterMissingError):
    code = "GCP_PARAMETER_MISSING_ERROR"


class GCPCredentialsError(CoiledException):
    code = "CREDENTIALS_ERROR"


class AWSCredentialsParameterError(ParameterMissingError):
    code = "AWS_PARAMETER_MISSING_ERROR"


class RegistryParameterError(ParameterMissingError):
    code = "REGISTRY_PARAMETER_ERROR"


class ParseIdentifierError(CoiledException):
    code = "PARSE_IDENTIFIER_ERROR"

    def __init__(self, message, *args, **kwargs):
        self.message = message
        super().__init__(message, *args, **kwargs)


class AccountConflictError(CoiledException):
    code = "ACCOUNT_CONFLICT_ERROR"

    def __init__(
        self,
        unparsed_name: str,
        account_from_name: str,
        account: str,
        message: Optional[str] = None,
        **kwargs,
    ):
        self.unparsed_name = unparsed_name
        self.account_from_name = account_from_name
        self.account = account
        if not message:
            message = (
                "Two accounts were specified where only "
                f'one is valid. The value "{unparsed_name}" '
                f'contains the account prefix "{account_from_name}," '
                f'and the account "{account}" was also passed.'
            )
        super().__init__(message, **kwargs)


class InstanceTypeError(CoiledException):
    code = "INSTANCE_TYPE_ERROR"


class GPUTypeError(CoiledException):
    code = "GPU_TYPE_ERROR"


class ArgumentCombinationError(CoiledException):
    code = "ARGUMENT_COMBINATION_ERROR"


class CidrInvalidError(CoiledException):
    code = "CIDR_INVALID_ERROR"


class PortValidationError(CoiledException):
    code = "PORT_VALIDATION_ERROR"


class AccountFormatError(CoiledException):
    code = "ACCOUNT_FORMAT_ERROR"


class ApiResponseStatusError(CoiledException):
    code = "API_RESPONSE_STATUS_ERROR"


class NotFound(CoiledException):
    code = "NOT_FOUND"


class PermissionsError(CoiledException):
    code = "PERMISSIONS_ERROR"



"""Status display widgets for Coiled cluster instances.

Each widget implementation must have an ``update()`` instance method that takes two
arguments. The first argument is the JSON response from the declarative
``/api/v2/clusters/account/{account}/id/{cluster_id}`` endpoint, parsed into a
dictionary. The second argument is a list of `coiled._beta.states.State` instances
that contain the state changes of the cluster, scheduler, and workers.

The widget implementation is then responsible for drawing itself to the environment.
Current implementations support Jupyter Notebook-like contexts and IPython console
contexts via the ``_ipython_display_`` and ``_repr_mimebundle_`` methods.

Please avoid adding unconditional dependencies on rendering or widget libraries, so
that people who only want to work in scripts don't have to install lots of packages
they won't need.
"""
from .util import HAS_RICH, sniff_environment

EXECUTION_CONTEXT = sniff_environment()

__all__ = ["EXECUTION_CONTEXT", "HAS_RICH"]



import click

from ..core import diagnostics as core_diagnostics
from .utils import CONTEXT_SETTINGS


@click.command(
    context_settings=CONTEXT_SETTINGS,
    help="Run diagnostics related to your local settings and account",
)
def diagnostics():
    """Run diagnostics related to your local settings and account."""
    diagnostics_result = core_diagnostics()
    print(diagnostics_result)




# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-10T22:42:59+0000",
 "dirty": false,
 "error": null,
 "full-revisionid": "587734f3b1d4f22d1ff5dd5f6f7fa3d18bab8d9b",
 "version": "0.5.3.post0.dev13"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



import logging
from collections import Counter
from datetime import datetime
from enum import Enum
from typing import List, NamedTuple, Optional, Tuple

from coiled.utils import COILED_LOGGER_NAME

logger = logging.getLogger(COILED_LOGGER_NAME)

DISPLAY_DATETIME_FORMAT = "%H:%M:%S (%Z)"


class BaseStateEnum(str, Enum):
    """
    Abstract class intended to be used for all state Enums
    inherits from str for easy comparisons
    """


class ClusterStateEnum(BaseStateEnum):
    """
    Valid states for Clusters
    """

    pending = "pending"
    starting = "starting"
    scaling = "scaling"
    ready = "ready"
    stopping = "stopping"
    stopped = "stopped"
    error = "error"

    def awaiting_readiness(self):
        return self.value in ("pending", "starting", "scaling")


class ClusterInfraStateEnum(BaseStateEnum):
    """Valid states for ClusterInfra."""

    queued = "queued"
    creating = "creating"
    error = "error"
    created = "created"
    destroying = "destroying"
    destroyed = "destroyed"


class ProcessStateEnum(BaseStateEnum):
    """
    Valid states for Processes
    """

    pending = "pending"
    starting = "starting"
    started = "started"
    stopping = "stopping"
    stopped = "stopped"
    error = "error"

    def awaiting_readiness(self):
        return self.value in ("pending", "starting", "scaling")


class InstanceStateEnum(BaseStateEnum):
    """
    Valid states for Instances
    """

    queued = "queued"
    starting = "starting"
    started = "started"
    ready = "ready"
    stopping = "stopping"
    stopped = "stopped"
    error = "error"


class StatefulObjectType(str, Enum):
    cluster = "cluster"
    scheduler = "scheduler"
    scheduler_instance = "scheduler_instance"
    cluster_infra = "cluster_infra"
    worker = "worker"
    worker_instance = "worker_instance"


class State(NamedTuple):
    type: StatefulObjectType
    updated: datetime
    name: str
    state: str
    reason: str


def flatten_log_states(types_to_states):
    states = []
    for type, one_type_states in types_to_states.items():
        for s in one_type_states:
            states.append(
                State(
                    type=StatefulObjectType(type),
                    updated=datetime.fromisoformat(s["updated"]),
                    name=s["stateful_object"]["name"] or "",
                    state=s["state"],
                    reason=s["reason"],
                )
            )

    states = sorted(states, key=lambda state: state.updated)
    return states


def log_states(states: List[State], level=None):
    """Log states, and return the latest one (for use in future queries)"""
    object_types_to_display = {
        StatefulObjectType.cluster: "Cluster",
        StatefulObjectType.scheduler: "Scheduler Process",
        StatefulObjectType.scheduler_instance: "Scheduler Instance",
        StatefulObjectType.cluster_infra: "Network Infrastructure",
        StatefulObjectType.worker: "Worker Process",
        StatefulObjectType.worker_instance: "Worker Instance",
    }
    max_type_length = max(len(v) for v in object_types_to_display.values())
    for s in states:
        updated_dt_str = datetime.strftime(s.updated.astimezone(), DISPLAY_DATETIME_FORMAT)
        log_string = (
            f"   | {object_types_to_display[s.type]:{max_type_length}} "
            f"| {s.name:46} | {s.state:10} at {updated_dt_str} | {s.reason}"
        )
        if s.state == "error":
            logger.log(msg=log_string, level=level or logging.ERROR)
        else:
            logger.log(msg=log_string, level=level or logging.DEBUG)


def get_process_instance_state(
    process: dict,
) -> Tuple[ProcessStateEnum, Optional[InstanceStateEnum]]:
    process_state = ProcessStateEnum(process["current_state"]["state"])
    instance = process["instance"]
    if instance is None:
        instance_state = None
    else:
        instance_state = InstanceStateEnum(instance["current_state"]["state"])
    return process_state, instance_state


class CombinedProcessStateEnum(BaseStateEnum):
    """This is a customer-facing type of status for a instance/process combo."""

    instance_queued = "queued"
    instance_starting = "starting"
    instance_running = "instance_running"  # instance is running; process not started or errored.
    downloading = "downloading"
    ready = "ready"
    stopping = "stopping"
    stopped = "stopped"
    error = "error"

    @classmethod
    def from_process_instance_states(cls, process_state, instance_state):
        if process_state == ProcessStateEnum.error or instance_state == InstanceStateEnum.error:
            return cls.error

        if instance_state is None:
            return cls.instance_queued  # a little bit of a lie
        elif instance_state == InstanceStateEnum.queued:
            return cls.instance_queued
        elif instance_state == InstanceStateEnum.starting:
            return cls.instance_starting
        elif instance_state == InstanceStateEnum.started:
            return cls.instance_running
        elif instance_state == InstanceStateEnum.ready:
            if process_state == ProcessStateEnum.started:
                return cls.ready
            elif process_state == ProcessStateEnum.starting:
                return cls.downloading
            elif process_state in [ProcessStateEnum.stopping, ProcessStateEnum.stopped]:
                return cls.stopping
            else:
                raise ValueError(f"unexpected instance/process state combo: {process_state}, {instance_state}")
        elif instance_state == InstanceStateEnum.stopping:
            return cls.stopping
        elif instance_state == InstanceStateEnum.stopped:
            return cls.stopped
        else:
            raise ValueError(f"unexpected instance state {instance_state}")


def get_combined_process_state(process) -> CombinedProcessStateEnum:
    """
    Collapses the process/instance state matrix to a single user-facing state summary.

    This necessarily elides some possible complexity and even lies a little.
    """
    process_state, instance_state = get_process_instance_state(process)

    return CombinedProcessStateEnum.from_process_instance_states(process_state, instance_state)


def summarize_status(details):
    combined_state_to_description = {
        CombinedProcessStateEnum.instance_queued: "queued for launch",
        CombinedProcessStateEnum.instance_starting: "starting",
        CombinedProcessStateEnum.instance_running: "running but not ready",
        CombinedProcessStateEnum.downloading: "downloading software environment",
        CombinedProcessStateEnum.ready: "ready",
        CombinedProcessStateEnum.stopping: "stopping",
        CombinedProcessStateEnum.stopped: "stopped",
        CombinedProcessStateEnum.error: "error",
    }
    max_description_length = max(len(desc) for desc in combined_state_to_description.values())

    cluster_state = details["current_state"]
    if cluster_state["state"] in [ClusterStateEnum.pending, ClusterStateEnum.starting]:
        return "Ensuring network infrastructure is ready..."
    else:
        scheduler = details["scheduler"]
        worker_state_counts = Counter(get_combined_process_state(worker) for worker in details["workers"])
        scheduler_status_to_show = f"Scheduler: {combined_state_to_description[get_combined_process_state(scheduler)]}"
        worker_status_descriptions = [
            f"{worker_state_counts[status]} {combined_state_to_description[status]}"
            for status in CombinedProcessStateEnum
            if worker_state_counts[status] > 0
        ]
        worker_statuses_to_show = (
            f"Workers: {', '.join(worker_status_descriptions)} " f"(of {sum(worker_state_counts.values())})"
        )
        return (
            f"{scheduler_status_to_show:{max_description_length + len('Scheduler: ') + 3}} "
            f"{worker_statuses_to_show}"
        )



import pytest

from coiled.errors import ServerError
from coiled.websockets import WebsocketConnector


@pytest.mark.asyncio
async def test_unhandled_websocket_exception(settings, cloud):
    settings.FF_SUPPRESS_UNKNOWN_EXCEPTIONS = True
    ws_server = cloud.server.replace("http", "ws", 1)
    ws_endpoint = f"{ws_server}/ws/api/v1/testing/exception/"
    session = cloud._ensure_session()
    ws = WebsocketConnector(
        endpoint=ws_endpoint,
        notifications_endpoint=f"{ws_server}/ws/api/v1/{cloud._default_account}/notifications/",
        session=session,
    )
    await ws.connect()
    await ws.send_json({"test": "test"})
    with pytest.raises(ServerError) as server_error:
        await ws.stream_messages()
    assert (
        "Coiled cloud encountered an unknown issue handling your request, contact customer service and quote ID".lower()
        in str(server_error.value).lower()
    )


@pytest.mark.asyncio
async def test_unhandled_websocket_exception_ff_off(settings, cloud):
    settings.FF_SUPPRESS_UNKNOWN_EXCEPTIONS = False
    ws_server = cloud.server.replace("http", "ws", 1)
    ws_endpoint = f"{ws_server}/ws/api/v1/testing/exception/"
    session = cloud._ensure_session()
    ws = WebsocketConnector(
        endpoint=ws_endpoint,
        notifications_endpoint=f"{ws_server}/ws/api/v1/{cloud._default_account}/notifications/",
        session=session,
    )
    await ws.connect()
    await ws.send_json({"test": "test"})
    with pytest.raises(ServerError) as server_error:
        await ws.stream_messages()
    assert "cloud.consumers.testing.WebsocketExceptionTest: Test".lower() in str(server_error.value).lower()


@pytest.mark.asyncio
async def test_api_websocket_exception(settings, cloud):
    session = cloud._ensure_session()
    ws_server = cloud.server.replace("http", "ws", 1)
    ws_endpoint = f"{ws_server}/ws/api/v1/testing/api-exception/"
    session = cloud._ensure_session()
    ws = WebsocketConnector(
        endpoint=ws_endpoint,
        notifications_endpoint=f"{ws_server}/ws/api/v1/{cloud._default_account}/notifications/",
        session=session,
    )
    await ws.connect()
    await ws.send_json({"test": "test"})
    with pytest.raises(ServerError) as server_error:
        await ws.stream_messages()
    assert "this message should be shown to user" == str(server_error.value)



import dask
import yaml
from click.testing import CliRunner

import coiled
from coiled.cli import env


def test_create_list_delete(sample_user):
    # No software environments initially
    assert not coiled.list_software_environments(new_build_backend=False)

    runner = CliRunner()
    name = "my-env"
    # Coiled env create creates a software environment
    result = runner.invoke(
        env.create,
        args=f"--name {name} --no-new-build-backend --container daskdev/dask:latest",
    )
    assert result.exit_code == 0

    result = coiled.list_software_environments(new_build_backend=False)
    assert f"{sample_user.account.name}/{name}" in result
    assert result[f"{sample_user.account.name}/{name}"]["container"] == "daskdev/dask:latest"

    # Coiled env list has similar output to coiled.list_software_environments
    result = runner.invoke(env.list, args="--no-new-build-backend", color=False)
    assert result.exit_code == 0
    output = yaml.safe_load(result.output)
    assert f"{sample_user.account.name}/{name}" in output
    assert output[f"{sample_user.account.name}/{name}"]["container"] == "daskdev/dask:latest"

    # Coiled delete removes software environments
    result = runner.invoke(env.delete, args=f"--no-new-build-backend {name}")
    assert result.exit_code == 0
    assert not coiled.list_software_environments(new_build_backend=False)


def test_inspect(sample_user, second_user, remote_access_url, sample_user_token):
    # Coiled inspect outputs software spec

    # Create software environment in sample_user's account
    with dask.config.set(
        {
            "coiled": {
                "user": f"{sample_user.user.username}",
                "token": sample_user_token,
                "server": remote_access_url,
                "account": sample_user.account.name,
                "backend-options": {"region": "us-east-2"},
                "no-minimum-version-check": True,
            }
        }
    ):
        coiled.create_software_environment(name="my-env", container="daskdev/dask:latest", new_build_backend=False)
        runner = CliRunner()
        result = runner.invoke(env.inspect, args=f"{sample_user.account.name}/my-env")
        assert result.exit_code == 0
        assert "daskdev/dask:latest" in result.output


def test_env_create(mocker):
    create_software_environment = mocker.patch("coiled.cli.env.create_software_environment")

    runner = CliRunner()
    result = runner.invoke(
        env.create,
        args=(
            "--name my-env "
            "--container daskdev/dask:latest "
            "--account my-account "
            "--environ ENV_VAR=VALUE "
            "--force-rebuild "
            "--private "
            "--no-new-build-backend "
        ),
    )

    assert result.exit_code == 0
    create_software_environment.assert_called_with(
        name="my-env",
        new_build_backend=False,
        container="daskdev/dask:latest",
        private=True,
        force_rebuild=True,
        environ={"ENV_VAR": "VALUE"},
        account="my-account",
        # Default values that we didn't change
        conda=None,
        pip=None,
        post_build=None,
        conda_env_name=None,
    )



from __future__ import annotations

import asyncio
import contextlib
import functools
import itertools
import json
import logging
import os
import random
import re
import ssl
import string
import subprocess
import sys
import tempfile
import threading
import time
import warnings
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from hashlib import md5
from logging.config import dictConfig
from math import ceil
from pathlib import Path
from typing import Dict, List, NoReturn, Optional, Tuple, Union
from urllib.parse import unquote, urlencode, urlparse
from zipfile import PyZipFile

import backoff
from packaging.version import Version

if sys.version_info >= (3, 8):
    from typing import Type, TypedDict, get_args, get_origin, get_type_hints
else:
    from typing_extensions import Type, TypedDict, get_args, get_origin, get_type_hints

import aiohttp
import boto3
import click
import dask
import rich
import urllib3
import yaml
from dask.distributed import Security
from rich.console import Console

from coiled.exceptions import (
    AccountFormatError,
    ApiResponseStatusError,
    CidrInvalidError,
    GPUTypeError,
    InstanceTypeError,
    ParseIdentifierError,
    PortValidationError,
    UnsupportedBackendError,
)

from .compatibility import COILED_VERSION, PY_VERSION
from .errors import ServerError

logger = logging.getLogger(__name__)


ACCOUNT_REGEX = re.compile(r"^[a-z0-9]+(?:[-_][a-z0-9]+)*$")
ALLOWED_PROVIDERS = ["aws", "vm_aws", "gcp", "vm_gcp"]

COILED_LOGGER_NAME = "coiled"
COILED_SERVER = "https://cloud.coiled.io"
COILED_RUNTIME_REGEX = re.compile(r"^coiled/coiled-runtime-(?P<version>\d+\-\d+\-\d+)-*")

# Dots after family names ensure we do not match t3a or other variants
AWS_BALANCED_INSTANCES_FAMILY_FILTER = ("t2.", "t3.", "m5.", "m6i.")
AWS_INSTANCES_FAMILY_FILTER = AWS_BALANCED_INSTANCES_FAMILY_FILTER  # deprecated
AWS_GPU_INSTANCE_FAMILIES_FILTER = ("g4dn",)
AWS_UNBALANCED_INSTANCE_FAMILIES_FILTER = ("c5.", "c6i.", "r6i.")

GCP_SCHEDULER_GPU = {
    "scheduler_accelerator_type": "nvidia-tesla-t4",
    "scheduler_accelerator_count": 1,
}


class VmType(TypedDict):
    """
    Example:
        {
        'name': 't2d-standard-8',
        'cores': 8,
        'gpus': 0,
        'gpu_name': None,
        'memory': 32768,
        'backend_type': 'vm_gcp'
        }
    """

    name: str
    cores: int
    gpus: int
    gpu_name: str
    memory: int
    backend_type: str


# TODO: copied from distributed, introduced in 2021.12.0.
# We should be able to remove this someday once we can increase
# the minimum supported version.
def in_async_call(loop, default=False):
    """Whether this call is currently within an async call"""
    try:
        return loop.asyncio_loop is asyncio.get_running_loop()
    except RuntimeError:
        # No *running* loop in thread. If the event loop isn't running, it
        # _could_ be started later in this thread though. Return the default.
        if not loop.asyncio_loop.is_running():
            return default
        return False


def validate_account(account: str):
    if ACCOUNT_REGEX.match(account) is None:
        raise AccountFormatError(
            f"Bad account format. Account '{account}' should be a combination "
            "of lowercase letters, numbers and hyhens."
        )


def random_str(length: int = 8):
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=length))


class GatewaySecurity(Security):
    """A security implementation that temporarily stores credentials on disk.

    The normal ``Security`` class assumes credentials already exist on disk,
    but our credentials exist only in memory. Since Python's SSLContext doesn't
    support directly loading credentials from memory, we write them temporarily
    to disk when creating the context, then delete them immediately."""

    def __init__(self, tls_key, tls_cert, extra_conn_args: Optional[dict] = None):
        self.tls_key = tls_key
        self.tls_cert = tls_cert
        self.extra_conn_args = extra_conn_args or {}

    def __repr__(self):
        return "GatewaySecurity<...>"

    def get_connection_args(self, role):
        ctx = None
        if self.tls_key and self.tls_cert:
            with tempfile.TemporaryDirectory() as tempdir:
                key_path = os.path.join(tempdir, "dask.pem")
                cert_path = os.path.join(tempdir, "dask.crt")
                with open(key_path, "w") as f:
                    f.write(self.tls_key)
                with open(cert_path, "w") as f:
                    f.write(self.tls_cert)
                ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=cert_path)
                ctx.verify_mode = ssl.CERT_REQUIRED
                ctx.check_hostname = False
                ctx.load_cert_chain(cert_path, key_path)
        return {
            "ssl_context": ctx,
            "require_encryption": True,
            "extra_conn_args": self.extra_conn_args,
        }


async def handle_api_exception(response, exception_cls: Type[Exception] = ServerError) -> NoReturn:
    with contextlib.suppress(aiohttp.ContentTypeError, json.JSONDecodeError, AttributeError):
        # First see if it's an error we created that has a more useful
        # message
        error_body = await response.json()
        if isinstance(error_body, dict):
            errs = error_body.get("non_field_errors")
            if "message" in error_body:
                raise exception_cls(error_body["message"])
            elif errs and isinstance(errs, list) and len(errs):
                raise exception_cls(errs[0])
            else:
                raise exception_cls("Server error, ".join(f"{k}={v}" for (k, v) in error_body.items()))
        else:
            raise exception_cls(error_body)

    error_text = await response.text()

    if not error_text:
        # Response contains no text/body, let's not raise an empty exception
        error_text = f"{response.status} - {response.reason}"
    raise exception_cls(error_text)


def normalize_server(server: Optional[str]) -> str:
    if not server:
        server = COILED_SERVER
    # Check if using an older server
    if "beta.coiledhq.com" in server or "beta.coiled.io" in server:
        # NOTE: https is needed here as http is not redirecting
        server = COILED_SERVER

    # remove any trailing slashes
    server = server.rstrip("/")

    return server


def get_account_membership(response: dict, account: Optional["str"] = None) -> Optional[dict]:
    if account is None:
        account = response.get("default_account")
        account = account.get("slug") if account else None

    membership_set = response.get("membership_set", [])

    for membership in membership_set:
        account_details = membership.get("account", {})
        has_membership = account_details.get("slug") == account
        if has_membership:
            return membership

    else:
        return None


def get_auth_header_value(token: str) -> str:
    """..."""
    # TODO: delete the branching after client only supports ApiToken.
    if "-" in token:
        return "ApiToken " + token
    else:
        return "Token " + token


def has_program_quota(membership: dict) -> bool:
    account_details = membership.get("account", {})
    return account_details.get("active_program", {}).get("has_quota") is True


def login_if_required(
    *,
    server: Optional[str] = None,
    token: Optional[str] = None,
    account: Optional[str] = None,
    save: Optional[bool] = None,
    use_config: bool = True,
    retry: bool = True,
):
    # "save" bool means always/never, None means try to do the thing that makes sense
    if save is None:
        # if token is already set in config and isn't being changed, then no need to save again
        if dask.config.get("coiled.token", None) and (not token or token == dask.config.get("coiled.token")):
            save = False
        else:
            # user doesn't already have token saved, so save it without asking
            save = True

    if use_config:
        token = token or dask.config.get("coiled.token")
        server = server or dask.config.get("coiled.server")
        if "://" not in server:
            server = "http://" + server
        server = server.rstrip("/")
        account = account or dask.config.get("coiled.account")

    try:
        asyncio.get_event_loop().run_until_complete(
            handle_credentials(server=server, token=token, account=account, save=save, retry=retry)
        )
    except ImportError as e:
        rich.print(f"[red]{e}")


# We only use ApiResponseStatusError on status 502, 503, 504
@backoff.on_exception(backoff.expo, ApiResponseStatusError, logger=logger)
async def handle_credentials(
    *,
    server: Optional[str] = None,
    token: Optional[str] = None,
    account: Optional[str] = None,
    save: Optional[bool] = None,
    retry: bool = True,
) -> Tuple[str, str, str]:
    """Validate and optionally save credentials

    Parameters
    ----------
    server
        Server to connect to. If not specified, will check the
        ``coiled.server`` configuration value.
    token
        Coiled user token to use. If not specified, will prompt user
        to input token.
    account
        Name of Coiled account to use. If not provided, will
        default to the user's default account.
    save
        Whether or not save credentials to coiled config file.
        If ``None``, will ask for input on whether or not credentials
        should be saved. Defaults to None.
    retry
        Whether or not to try again if invalid credentials are entered.
        Retrying is often desired in interactive situations, but not
        in more programmatic scenerios. Defaults to True.

    Returns
    -------
    user
        Username
    token
        User API token
    server
        Server being used
    """
    if account:
        validate_account(account)
    # If testing locally with `ngrok` we need to
    # rewrite the server to localhost
    server = server or dask.config.get("coiled.server", COILED_SERVER)
    server = normalize_server(server)
    login_url = f"{server}/profile"
    if token is None:
        rich.print(f"Please login to [link]{login_url}[/link] to get your token")
        # Using click instead of getpass to make testing easier
        token = click.prompt("Token", hide_input=True)
        if not token:
            raise ValueError(
                "Token not provided, unable to login. You can create new tokens and "
                f"managing your existing ones at {server}/profile.\n"
            )

    # TODO: revert when we remove versioneer
    if dask.config.get("coiled.no-minimum-version-check", False):
        client_version = "coiled-frontend-js"
    else:
        client_version = COILED_VERSION

    # Validate token and get username
    async with aiohttp.ClientSession(
        headers={
            "Authorization": get_auth_header_value(token),
            "Client-Version": client_version,
        }
    ) as session:
        response = await session.request("GET", f"{server}/api/v1/users/me/")

        try:
            data = await response.json()
        except (aiohttp.ContentTypeError, json.JSONDecodeError):
            # The code bellow will handle status codes that are errors,
            # we are setting data to an empty dict to make typing happy
            data = {}

        if response.status == 426:
            # Upgrade required
            await handle_api_exception(response)
        elif response.status in [502, 503, 504]:
            raise ApiResponseStatusError(
                f"Unable to receive data from the server. Received {response.status} - Temporary Error."
            )
        elif response.status >= 400:
            if isinstance(data, dict) and retry and "Invalid token" in data.get("detail", ""):
                # Token is not valid, ask for a new one
                rich.print(
                    "[red]Invalid Coiled token encountered. You can create new tokens and manage "
                    f"your existing ones at {server}/profile.\n"
                )
                return await handle_credentials(server=server, token=None, save=None)
            else:
                await handle_api_exception(response)

        account_membership = get_account_membership(data, account=account)
        # only validate if account arg is provided by user
        if account and not account_membership:
            rich.print("[red]You are not a member of this account. Perhaps try another one?\n")
            account = click.prompt("Account")
            if account:
                validate_account(account)
            else:
                rich.print("[red]No account provided, unable to login.")

            return await handle_credentials(server=server, token=token, save=None, account=account)

    # We should always have username from above, but let's be defensive about it just in case.
    user = data.get("username")
    if not user:
        raise ValueError(
            "Unable to get your account username after login. Please try to login again, if "
            "the error persists, please reach out to Coiled Support at support@coiled.io"
        )

    # Avoid extra printing when creating clusters
    if save is not False:
        rich.print("[green]Authentication successful")
        if not account_membership or not has_program_quota(account_membership):
            rich.print("[red]You have reached your quota of Coiled credits for this account.")
    if save is None:
        # Optionally save user credentials for next time
        save_creds = input("Save credentials for next time? [Y/n]: ")
        if save_creds.lower() in ("y", "yes", ""):
            save = True
    if save:
        config_file = os.path.join(os.path.expanduser("~"), ".config", "dask", "coiled.yaml")
        # Make sure directory with config exists
        os.makedirs(os.path.dirname(config_file), exist_ok=True)
        configs = dask.config.collect_yaml([config_file])
        creds = {
            "coiled": {
                "user": user,
                "token": token,
                "server": server,
                "account": account_membership.get("account", {}).get("slug") if account_membership else None,
            }
        }
        config = dask.config.merge(*configs, creds)
        with open(config_file, "w") as f:
            f.write(yaml.dump(config))
        rich.print(f"Credentials have been saved at [blue]{config_file}")
        # Make sure saved configuration values are set for the current Python process
        dask.config.set(config)

    return user, token, server


class Spinner:
    """A spinner context manager to denotate we are still working"""

    def __init__(self, delay=0.2):
        self.spinner = itertools.cycle(["-", "/", "|", "\\"])
        self.delay = delay
        self.busy = False

    def write_next(self):
        with self._screen_lock:
            sys.stdout.write(next(self.spinner))
            sys.stdout.flush()

    def remove_spinner(self, cleanup=False):
        with self._screen_lock:
            sys.stdout.write("\b")
            if cleanup:
                sys.stdout.write(" ")  # overwrite spinner with blank
                sys.stdout.write("\r")  # move to next line
            sys.stdout.flush()

    def spinner_task(self):
        while self.busy:
            self.write_next()
            time.sleep(self.delay)
            self.remove_spinner()

    def __enter__(self):
        if sys.stdout.isatty():
            self._screen_lock = threading.Lock()
            self.busy = True
            self.thread = threading.Thread(target=self.spinner_task)
            self.thread.start()

    def __exit__(self, exception, value, tb):
        if sys.stdout.isatty():
            self.busy = False
            self.remove_spinner(cleanup=True)
        else:
            sys.stdout.write("\r")


def parse_identifier(
    identifier: str,
    property_name: str = "name",
    can_have_revision: bool = False,
    allow_uppercase: bool = False,
) -> Tuple[Optional[str], str, Optional[str]]:
    """
    Parameters
    ----------
    identifier:
        identifier of the resource, i.e. "coiled/xgboost" "coiled/xgboost:1ef489", "xgboost:1ef489" or "xgboost"
    can_have_revision:
        Indicates if this identifier supports having a ":<string>" postfix, as in
        the ":1ef489" in "xgboost:1ef489". At time of writing, this only has an effect
        on software environments. For other resources this has no meaning. At time
        of writing, this only affects the error message that will be printed out.
    property_name:
        The name of the parameter that was being validated; will be printed
        with any error messages, i.e. Unsupported value for "software_environment".

    Examples
    --------
    >>> parse_identifier("coiled/xgboost", "software_environment")
    ("coiled", "xgboost", None)
    >>> parse_identifier("xgboost", "software_environment", False)
    (None, "xgboost", None)
    >>> parse_identifier("coiled/xgboost:1ef4543", "software_environment", True)
    ("coiled", "xgboost", "1ef4543")

    Raises
    ------
    ParseIdentifierError
    """
    if allow_uppercase:
        rule = re.compile(r"^([a-zA-Z0-9-_]+?/)?([a-zA-Z0-9-_]+?)(:[a-zA-Z0-9-_]+)?$")
        rule_text = ""
    else:
        rule = re.compile(r"^([a-z0-9-_]+?/)?([a-z0-9-_]+?)(:[a-zA-Z0-9-_]+)?$")
        rule_text = "lowercase "

    match = re.fullmatch(rule, identifier)
    if match:
        account, name, revision = match.groups()
        account = account.replace("/", "") if account else account
        revision = revision.replace(":", "") if revision else revision
        return account, name, revision

    if can_have_revision:
        message = (
            f"Invalid {property_name}: should have format (<account>/)<name>(:<revision>),"
            ' for example "coiled/xgboost:1efd456", "xgboost:1efd456" or "xgboost". '
            f"It can only contain {rule_text}ASCII letters, numbers, hyphens and underscores. "
            "The <name> is required (xgboost in the previous example). The <account> prefix"
            f" can be used to specify a {property_name} from a different account, and the"
            f" :<revision> can be used to select a specific revision of the {property_name}."
        )
    else:
        message = (
            f"Invalid {property_name}: should have format (<account>/)<name>,"
            f' for example "coiled/xgboost" or "python-37". and can only contain {rule_text}ASCII letters,'
            ' numbers, hyphens and underscores. The <name> is required ("xgboost" and "python-37"'
            " in the previous examples)."
            f' The <account> prefix (i.e. "coiled/") can be used to specify a {property_name}'
            " from a different account."
        )
    raise ParseIdentifierError(message)


def get_platform():
    # Determine platform
    if sys.platform == "linux":
        platform = "linux"
    elif sys.platform == "darwin":
        platform = "osx"
    elif sys.platform == "win32":
        platform = "windows"
    else:
        raise ValueError(f"Invalid platform {sys.platform} encountered")
    return platform


class ExperimentalFeatureWarning(RuntimeWarning):
    """Warning raise by an experimental feature"""

    pass


class DeprecatedFeatureWarning(RuntimeWarning):
    pass


def experimental(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        warnings.warn(
            f"{func.__name__} is an experimental feature which is subject "
            "to breaking changes, being removed, or otherwise updated without notice "
            "and should be used accordingly.",
            ExperimentalFeatureWarning,
        )
        return func(*args, **kwargs)

    return inner


async def run_command_in_subprocess(cmd: str, suppress_error: bool = False):
    """Run the command in subprocess.

    If a user is using a python version < 3.8 then he will get a RuntimeError exception -
    ``Cannot add child handler, the child watcher does not have a loop attached``. As a
    workaround for the issue, we check if the user is running any earlier version of python
    and run the command in subprocess instead.

    For reference on the RuntimeError exception: https://bugs.python.org/issue35621

    """
    platform = get_platform()

    if PY_VERSION < "3.8" or platform == "windows":
        proc = subprocess.Popen(
            cmd.split(" "),
            bufsize=1,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.stdout
        for line in proc.stdout:
            yield line.rstrip()

    else:
        proc = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        lines = []
        while proc.stdout and proc.stderr and proc.returncode is None:
            await asyncio.sleep(0.5)
            async for line in proc.stdout:
                line = line.decode().rstrip()
                lines.append(line)
                yield line

            async for line in proc.stderr:
                line = line.decode().rstrip()
                lines.append(line)
                yield line

        if proc.returncode and not suppress_error:
            raise ValueError("\n".join(lines))


def name_exists_in_dict(user: Union[str, None], name: Union[str, None], dictionary: Union[dict, None]) -> bool:
    """Check if name exists in dictionary.

    Since we are returning ``<account>/<name>`` when listing software envs
    and cluster configs, this method will try to handle the case where sometimes
    we are passing a name with the format ``<account>/<name>`` and sometimes with
    ``<name``.

    If we can't find ``<account>/<name>`` in the dictionary, we will check if we
    can find the ``name`` substring in the dictionary keys.

    """
    if dictionary:
        if f"{user}/{name}" in dictionary or name in dictionary:
            return True
        if any([soft_env for soft_env in dictionary if name in soft_env]):
            return True
    return False


def rich_console():
    is_spyder = False

    with contextlib.suppress(AttributeError, ImportError):
        from IPython import get_ipython

        ipython = get_ipython().config
        if ipython.get("SpyderKernelApp"):
            is_spyder = True

    if is_spyder:
        print("Creating Cluster. This might take a few minutes...")
        return Console(force_jupyter=False)
    return Console()


def verify_aws_credentials(aws_access_key_id: str, aws_secret_access_key: str):
    """Verifies AWS Credentials are valid"""
    sts = boto3.client(
        "sts",
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    sts.get_caller_identity()


def scheduler_ports(protocol: Union[str, List[str]]):
    """Generate scheduler ports based on protocol(s)"""
    exclude = 8787  # dashboard port
    start = 8786
    if isinstance(protocol, str):
        return start

    port = start
    ports = []
    for _ in protocol:
        if port == exclude:
            port += 1
        ports.append(port)
        port += 1
    return ports


def parse_gcp_region_zone(region: Optional[str] = None, zone: Optional[str] = None):
    """Parse GCP zone and region or return default region/zone.

    This is an helper function to make it easier for us
    to parse gcp zones. Since users can specify regions,
    zones or one of the two, we should create some sane
    way to deal with the different combinations.

    """
    if not region and not zone:
        region = "us-east1"
        zone = "us-east1-c"
    elif region and zone and len(zone) == 1:
        zone = f"{region}-{zone}"
    elif not region and zone and len(zone) == 1:
        region = "us-east1"
        zone = f"{region}-{zone}"
    elif zone and not region:
        region = zone[:-2]

    return region, zone


class UTCFormatter(logging.Formatter):
    converter = time.gmtime


def enable_debug_mode():
    from .context import COILED_SESSION_ID

    LOGGING = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "utc": {
                "()": UTCFormatter,
                "format": "[UTC][%(asctime)s][%(levelname)-8s][%(name)s] %(message)s",
            },
        },
        "handlers": {
            "utc-console": {
                "class": "logging.StreamHandler",
                "formatter": "utc",
            },
        },
        "loggers": {
            "coiled": {
                "handlers": ["utc-console"],
                "level": logging.DEBUG,
            },
        },
    }
    dictConfig(LOGGING)
    logger.debug(f"Coiled Version : {COILED_VERSION}")
    start = datetime.now(tz=timezone.utc)
    trace_link = get_datadog_trace_link(start=start, **{"coiled-session-id": COILED_SESSION_ID})
    logger.info(f"DD Trace: {trace_link}")
    trace_link = get_datadog_logs_link(start=datetime.now(tz=timezone.utc), **{"coiled-session-id": COILED_SESSION_ID})
    logger.info(f"DD Logs: {trace_link}")


def get_datadog_trace_link(
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    **filters: Dict[str, str],
):
    params = {
        "query": " ".join([f"@{k}:{v}" for k, v in filters.items()]),
        "paused": "true",
        "streamTraces": "false",
        "showAllSpans": "true",
    }
    if start:
        fuzzed = start - timedelta(minutes=1)
        params["start"] = str(int(fuzzed.timestamp() * 1000))
    if end:
        fuzzed = end + timedelta(minutes=1)
        params["end"] = str(int(fuzzed.timestamp() * 1000))
    return f"https://app.datadoghq.com/apm/traces?{urlencode(params)}"


def get_datadog_logs_link(
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    **filters: Dict[str, str],
):
    params = {
        "query": " ".join([f"@{k}:{v}" for k, v in filters.items()]),
        "live": "false",
    }
    if start:
        fuzzed = start - timedelta(minutes=1)
        params["from_ts"] = str(int(fuzzed.timestamp() * 1000))
    if end:
        fuzzed = end + timedelta(minutes=1)
        params["to_ts"] = str(int(fuzzed.timestamp() * 1000))
    return f"https://app.datadoghq.com/logs?{urlencode(params)}"


def validate_gpu_type(gpu_type: str):
    """Validate gpu type provided by the user.

    Currently, we are just accepting the nvidia-tesla-t4 gpu type
    but in the next iteration we will accept more types. We are also
    filtering out virtual workstation gpus upstream, so we need this
    function to remove types that we know it will fail.

    """

    if gpu_type != "nvidia-tesla-t4":
        error_msg = f"GPU type '{gpu_type}' is not a supported GPU type. Allowed " "GPU types are: 'nvidia-tesla-t4'."
        raise GPUTypeError(error_msg)
    return True


def is_gcp(account: str, accounts: dict) -> bool:
    """Determine if an account backend is Google Cloud Provider.

    Parameters
    ----------
    account: str
        Slug of Coiled account to use.
    accounts: dict
        Dictionary of available accounts from the /api/v1/users/me endpoint.

    Returns
    -------
    gcp_backend: bool
        True if a GCP backend, else False.
    """
    user_account = accounts.get(account, {})
    user_options = user_account.get("options", {})
    gcp_backend = user_account.get("backend") == "vm_gcp" or user_options.get("provider_name") == "gcp"
    return gcp_backend


def is_customer_hosted(account: str, accounts: dict) -> bool:
    """Determine if an account backend is customer-hosted.

    Parameters
    ----------
    account: str
        Slug of Coiled account to use.
    accounts: dict
        Dictionary of available accounts from the /api/v1/users/me endpoint.

    Returns
    -------
    customer_hosted: bool
        True if a customer-hosted backend, else False.
    """
    user_account = accounts.get(account, {})
    backend_type = user_account.get("backend")
    customer_hosted = backend_type == "vm"
    return customer_hosted


def validate_cidr_block(cidr_block: str):
    """Validate cidr block added by the user.

    Here we are only checking if the cidr block is in the
    format <int>.<int>.<int>.<int>/<int>.

    """
    if not isinstance(cidr_block, str):
        raise CidrInvalidError(
            f"CIDR needs to be of type string, but received '{type(cidr_block)}' "
            "please specify your CIDR block as a string."
        )
    match = re.fullmatch(r"(\d{1,3}\.){3}\d{1,3}\/\d{1,3}", cidr_block)
    if not match:
        raise CidrInvalidError(
            f"The CIDR block provided '{cidr_block}' doesn't appear "
            "to have the correct IPV4 pattern, your CIDR block should "
            "follow the '0.0.0.0/0' pattern."
        )
    return True


def validate_ports(ports):
    """Validate the ports tha the user tries to pass.

    We need to make sure that the user passes a list of ints only,
    otherwise we should raise an exception.

    """
    if not isinstance(ports, list):
        raise PortValidationError(
            f"Ports need to be of type list, but received '{type(ports)}' " "please adds your ports in a list."
        )

    for port in ports:
        if not isinstance(port, int):
            raise PortValidationError(
                f"Ports need to be of type int, but received '{type(port)}' "
                "please use a int value for your port number."
            )
    return True


def validate_network(backend_options: dict, is_customer_hosted: bool):
    """Validate network configuration from backend options.

    Users can specify network and/or subnet(s) to use. We need to validate these and
    ensure that they're only used with customer hosted backend.
    """
    if not is_customer_hosted:
        raise UnsupportedBackendError(
            "Network configuration isn't available in a Coiled "
            "hosted backend. Please change your backend to configure network."
        )

    network_options = backend_options.get("network", {})

    # check that resource ids, if specified, are strings
    resource_ids = (
        network_options.get("network_id"),
        network_options.get("scheduler_subnet_id"),
        network_options.get("worker_subnet_id"),
        network_options.get("firewall_id"),
    )

    for resource_id in resource_ids:
        if resource_id and not isinstance(resource_id, str):
            raise TypeError(f"Network id '{id}' for '{resource_id}' must be a string.")


def parse_backend_options(
    backend_options: dict,
    account: str,
    accounts: dict,
    worker_gpu: Optional[int],
) -> dict:
    """Parse backend options before launching cluster.

    The following are checked and parsed before launching a cluster:
        - If launching into a GCP cluster, `preemptible` is aliased with `spot`.
        - If requesting worker GPUs, `spot` defaults to `False` unless specified.
        - If set, `zone` overrides `gcp_zone` (for GCP).

    Parameters
    ----------
    backend_options: dict
        Dictionary of backend specific options (e.g. ``{'region': 'us-east-2'}``).
    account: str
        Slug of Coiled account to use.
    accounts: dict
        Dictionary of available accounts from the /api/v1/users/me endpoint.
    worker_gpu: int
        Number of GPUs allocated for each worker.

    Returns
    -------
    backend_options: dict
        A dictionary of parsed backend options.
    """
    backend_options = deepcopy(backend_options)
    gcp_backend = is_gcp(account, accounts)
    customer_hosted = is_customer_hosted(account, accounts)
    # alias preemptible/spot
    if backend_options.get("preemptible") is not None and gcp_backend:
        backend_options["spot"] = backend_options.pop("preemptible")
    # default to on-demand for gpu workers
    if backend_options.get("spot") is None and bool(worker_gpu):
        backend_options["spot"] = False

    # zone (if set) overrides gcp_zone for gcp
    if backend_options.get("zone") and backend_options.get("gcp_project_name"):
        backend_options["gcp_zone"] = backend_options.get("zone")

    if backend_options.get("network"):
        validate_network(backend_options, customer_hosted)
    if backend_options.get("firewall"):
        cidr = backend_options["firewall"].get("cidr")
        ports = backend_options["firewall"].get("ports")
        if ports:
            validate_ports(ports)
        if cidr:
            validate_cidr_block(cidr)
    return backend_options


async def get_worker_creation_notification_error(cloud, account: Optional[str]) -> Optional[str]:
    """Call get_notifications and return error messages from worker creation.

    When we try to create workers, we use the ``create_worker`` event type to
    the notification, we can quickly use this event type and the logging level
    ``ERROR`` to only get notifications that are related to worker creation.

    Note that we don't use the cluster name in the backend, but the cluster id
    so we can't really filter the notifications by name, with that in mind, we
    make the assumption that the last notification that we receive will be the
    one from the cluster that just ran. This is ugly but should work okay for
    now.

    """
    notifications = await cloud.get_notifications(account=account, json=True, event_type="create_worker", level="ERROR")

    if notifications:
        last_notification = notifications[-1]
        error_message = last_notification.get("data", {}).get("error")

        # If there is no reason, lets avoiding showing "giving up" messages
        # with no useful message
        if error_message and "Reason" in error_message:
            return error_message
        return last_notification["msg"]


def parse_wait_for_workers(n_workers: int, wait_for_workers: Optional[bool | int | float] = None) -> int:
    """Parse the option to wait for workers."""
    wait_for_workers = (
        # Set 30% as default value to wait for workers
        dask.config.get("coiled.wait-for-workers", 0.3)
        if wait_for_workers is None
        else wait_for_workers
    )

    if wait_for_workers is True:
        to_wait = n_workers
    elif wait_for_workers is False:
        to_wait = 0
    elif isinstance(wait_for_workers, int):
        if wait_for_workers >= 0 and wait_for_workers <= n_workers:
            to_wait = wait_for_workers
        else:
            raise ValueError(
                f"Received invalid value '{wait_for_workers}' as wait_for_workers, "
                f"this value needs to be between 0 and {n_workers}"
            )
    elif isinstance(wait_for_workers, float):
        if wait_for_workers >= 0 and wait_for_workers <= 1.0:
            to_wait = ceil(wait_for_workers * n_workers)
        else:
            raise ValueError(
                f"Received invalid value '{wait_for_workers}' as wait_for_workers, "
                "this value needs to be a value between 0 and 1.0."
            )
    else:
        raise ValueError(
            f"Received invalid value '{wait_for_workers}' as wait_for_workers, "
            "this value needs to be either a Boolean, an Integer or a Float."
        )

    return to_wait


def bytes_to_mb(bytes: float) -> int:
    """Convert bytes to Mb.

    Our backend expects the memory to be passed as Mb, so we need to convert
    bytes obtained by dask to Mb.

    """
    return round(abs(bytes) * 10 ** -6)


def parse_requested_memory(
    memory: Optional[Union[int, str, float, list[str], list[int], list[float]]],
    min_memory: Optional[Union[str, int, float]],
) -> dict:
    """Handle memory requested by user.

    Users calling `list_instance_types` can use different ways to specify memory,
    this function will handle the different cases and return a dictionary with the
    expected result.

    Note: We are also giving a +- 10% buffer on the requested memory.

    """
    parsed_memory = {}

    if isinstance(memory, list):
        if len(memory) > 2:
            raise ValueError(
                f"Memory should contain only two values in the format `[minimum, maximum]`, but received {memory}."
            )
        parsed_memory["memory__gte"] = bytes_to_mb(dask.utils.parse_bytes(memory[0]) * 0.89)
        parsed_memory["memory__lte"] = bytes_to_mb(dask.utils.parse_bytes(memory[1]) * 1.1)
    elif isinstance(memory, int) or isinstance(memory, float) or isinstance(memory, str):
        parsed_memory["memory__gte"] = bytes_to_mb(dask.utils.parse_bytes(memory) * 0.89)
        parsed_memory["memory__lte"] = bytes_to_mb(dask.utils.parse_bytes(memory) * 1.1)

    if min_memory and not memory:
        parsed_memory["memory__gte"] = bytes_to_mb(dask.utils.parse_bytes(min_memory))

    return parsed_memory


def get_instance_type_from_cpu_memory(
    cpu: Optional[Union[int, List[int]]] = None,
    memory: Optional[Union[int, str, float, list[int], list[str], list[float]]] = None,
    gpus: Optional[int] = None,
    backend: Optional[str] = None,
) -> List[str]:
    """Get instances by cpu/memory combination.

    We are using the `list_instance_types` method to get the
    instances that match the cpu/memory specification. If no
    instance can be found we will raise an exception
    informing the user about this.

    """
    # Circular imports
    from .core import list_instance_types

    if backend == "gcp":
        instance_family = "n1" if gpus else "e2"
        instances = [
            instance_name
            for instance_name in list_instance_types(cores=cpu, memory=memory, backend=backend)
            if instance_name.startswith(instance_family)
        ]
    else:
        instance_types = list_instance_types(cores=cpu, memory=memory, backend=backend, gpus=gpus)

        aws_family_filter = AWS_BALANCED_INSTANCES_FAMILY_FILTER
        if gpus:
            aws_family_filter = AWS_GPU_INSTANCE_FAMILIES_FILTER
        elif cpu and memory:
            aws_family_filter += AWS_UNBALANCED_INSTANCE_FAMILIES_FILTER

        instances = [instance_name for instance_name in instance_types if instance_name.startswith(aws_family_filter)]

    if not instances:
        error_message = (
            "Unable to find instance types that match the specification: \n"
            f"\t Cores: {cpu}  Memory: {memory} GPUs: {gpus}\n"
        )

        if cpu or memory:
            error_message += (
                "You can try selecting a range for the cpu or memory, for example: "
                "`cpu=[2, 8]` or `memory=['32Gb','64Gb'] \n"
            )

        if cpu and memory:
            memory_matching_instances = [instance_name for instance_name in list_instance_types(memory=memory)]

            if memory_matching_instances:
                error_message += (
                    "You might want to pick these instances that match your "
                    "memory requirements, but have different core count. \n"
                    f"{memory_matching_instances} \n"
                )
        error_message += (
            f'You can also use `coiled.list_instance_types(backend="{backend}")` to list possible instance types '
            "and use the `scheduler_vm_types=[]` or `worker_vm_types=[]` keyword "
            "argument to specify the specific instance type you want."
        )
        raise InstanceTypeError(error_message)
    return instances


def validate_type(type_, obj_) -> bool:
    if hasattr(type_, "__forward_arg__"):
        # evaluate ForwardRef types -- this function breaks if we can't
        return validate_type(type_.__forward_arg__, obj_)
    try:
        if hasattr(type_, "__annotations__"):
            # check for type TypedDict

            # check that object is a dict
            if not isinstance(obj_, dict):
                return False

            # check all the keys
            for k, v in get_type_hints(type_).items():
                # get so that undefined key will be None, and check will
                # pass only if type is Optional
                if not validate_type(v, obj_.get(k)):
                    return False

            # make sure no extra keys
            obj_keys = set(obj_.keys())
            type_keys = set(get_type_hints(type_).keys())
            if len(obj_keys - type_keys):
                return False

            return True

        elif hasattr(type_, "__origin__"):
            if get_origin(type_) == Union:
                # check for type Union (including Optional)
                union_types = get_args(type_)
                return any((validate_type(subtype, obj_) for subtype in union_types))
            elif get_origin(type_) == list:
                # for list, check that each item has correct type
                subtype = get_args(type_)[0]
                return all((validate_type(subtype, list_item) for list_item in obj_))

        # check for other types (can check directly)
        return isinstance(obj_, type_)

    except Exception:
        # assume that this function works for the types we're checking
        return False


def validate_vm_typing(vm_types: Optional[List[str]]):
    """Validate instance typing.

    We need to add this function because the error that our API is returning
    isn't exactly user friendly. We should raise a type error with an informative
    message instead of the dictionary that the API throws.

    """
    if not vm_types:
        return
    if not isinstance(vm_types, list):
        raise TypeError(
            f"Instance types must be a list, but the value '{vm_types}' is of "
            f"type: {type(vm_types)}. Please use a list of strings when specifying "
            "instance types."
        )
    for instance in vm_types:
        if not isinstance(instance, str):
            raise TypeError(
                f" Instance types must be a string, but '{instance}' is of type "
                f"{type(instance)}. Please use a string instead."
            )


def name_to_version(name: str) -> Version:
    matched = COILED_RUNTIME_REGEX.match(name)
    if matched is None:
        return Version("0.0.0")  # Should not happen
    return Version(matched.group("version").replace("-", "."))


def get_details_url(server, account, cluster_id):
    if cluster_id is None:
        return None
    else:
        return f"{server}/clusters/{cluster_id}/overview?account={account}"


def get_grafana_url(
    cluster_details,
    account,
    cluster_id,
    grafana_server="grafana.dev-sandbox.coiledhq.com",
):
    cluster_name = cluster_details["name"]

    # for stopped clusters, only get metrics until the cluster stopped
    if cluster_details["current_state"]["state"] in ("stopped", "error"):
        end_ts = int(datetime.fromisoformat(cluster_details["current_state"]["updated"]).timestamp() * 1000)
    else:
        end_ts = "now"

    start_ts = int(datetime.fromisoformat(cluster_details["scheduler"]["created"]).timestamp() * 1000)

    base_url = f"https://{grafana_server}/d/eU1bT-nVz/cluster-metrics-prometheus?orgId=1"
    datasource = "default"  # FIXME
    # FIXME -- get account name for cluster

    full_url = (
        f"{base_url}"
        f"&datasource={datasource}"
        f"&var-env=All"
        f"&var-account={account}"
        f"&var-cluster={cluster_name}"
        f"&var-cluster-id={cluster_id}"
        f"&from={start_ts}"
        f"&to={end_ts}"
    )

    return full_url


def cluster_firewall(
    target: str,
    scheduler: int = 8786,
    dashboard: Optional[int] = 8787,
    jupyter: bool = False,
    ssh: bool = False,
    http: bool = False,
    https: bool = True,
):
    """
    Easier cluster firewall configuration when using a single CIDR.

    Examples
    --------
    To create a cluster than only accepts connections from your IP address,
    and opens port 22 for SSH access:
    >>> coiled.Cluster(
            ...,
            backend_options={**coiled.utils.cluster_firewall("me", ssh=True)}
        )

    Parameters
    ----------
    target: str
        Open cluster firewall to this range. You can either specify a CIDR,
        or use "me" to automatically get your IP address and use that, or
        "everyone" for "0.0.0.0/0".
    scheduler: int
        Port to open for access to scheduler, 8786 by default.
    dashboard: Optional[int]
        Port to open for access to dashboard, 8787 by default.
        Use `None` to not open a port for dashboard access.
    jupyter: bool
        Open port used for jupyter (8888), closed by default.
    ssh: bool
        Open port used for ssh (22), closed by default.
    http: bool
        Open port used for http (80), closed by default.
    https: bool
        Open port used for https (443), open by default.
    """
    if target == "everyone":
        cidr = "0.0.0.0/0"
    elif target == "me":
        # get public/internet routable address from which local user will be hitting scheduler
        with urllib3.PoolManager() as pool:
            my_public_ip = pool.request("GET", "https://api.ipify.org").data.decode("utf-8")

        cidr = f"{my_public_ip}/32"
    else:
        # TODO validate this this is cidr
        cidr = target

    ports = [scheduler]

    if dashboard:
        ports.append(dashboard)

    if ssh:
        ports.append(22)
    if jupyter:
        ports.append(8888)
    if http:
        ports.append(80)
    if https:
        ports.append(443)
    return {"ingress": [{"ports": ports, "cidr": cidr}]}


async def validate_wheel(wheel: Path) -> Tuple[bool, str]:
    """
    Validate a wheel contains some python files and return
    the md5
    """
    hash = md5()
    has_python = False
    with PyZipFile(str(wheel), mode="r") as wheelzip:
        info = wheelzip.infolist()
        for file in info:
            if not has_python and file.filename != "__init__.py" and file.filename.endswith(".py"):
                has_python = True
            if "dist-info" not in file.filename:
                hash.update(str(file.CRC).encode())
    return has_python, hash.hexdigest()


def get_aws_identity():
    import boto3
    import botocore

    response = {}

    session = boto3.Session()
    sts = session.client("sts")

    def get_account_alias() -> str:
        try:
            iam = session.client("iam")
            r = iam.list_account_aliases()
            return r.get("AccountAliases", [])[0]
        except Exception:
            return ""

    try:
        identity = sts.get_caller_identity()

        response["account"] = identity.get("Account")
        response["who"] = identity.get("Arn").split(":")[-1]

        alias = get_account_alias()
        if alias:
            response["account_alias"] = alias
    except botocore.exceptions.NoCredentialsError:
        response["error"] = "NoCredentialsError"
    except Exception as e:
        response["error"] = str(e)

    return response


def parse_file_uri(file_uri: str) -> Path:
    p = urlparse(file_uri)
    unquoted_path = unquote(p.path)
    if os.name == "nt":
        # handling valid windows URIs of
        # file:///C:/somewhere/something.whl
        unquoted_path = unquoted_path.lstrip("/")
    return Path(unquoted_path)



import traceback

import click
from rich import print
from rich.prompt import Prompt

import coiled

from ..utils import CONTEXT_SETTINGS
from .util import setup_failure


@click.command(context_settings=CONTEXT_SETTINGS)
def setup_wizard() -> bool:
    return do_setup_wizard()


def do_setup_wizard() -> bool:
    coiled.utils.login_if_required()

    coiled.add_interaction(
        action="cli-setup-wizard:start",
        success=True,
    )

    print(
        "Welcome to Coiled!\n\n"
        "To begin we need to connect Coiled to your cloud account.\n"
        "Select one of the following options:\n\n"
        "  1. Amazon Web Service (AWS)\n"
        "  2. Google Cloud Platform (GCP)\n"
        "  3. I don't have a cloud account, set me up with a free trial?\n"
        "  [red]x[/red]. Exit setup\n"
    )

    try:
        choice = Prompt.ask(
            "Choice",
            choices=["1", "2", "3", "x"],
            show_choices=False,
        )
    except KeyboardInterrupt:
        coiled.add_interaction(action="cli-setup-wizard:KeyboardInterrupt", success=False)
        return False

    coiled.add_interaction(action="cli-setup-wizard:prompt", success=True, choice=choice)

    if choice == "1":  # AWS
        print("\nRunning [green]coiled setup aws[/green]\n")
        from .aws import do_setup

        try:
            return do_setup(slug="coiled")
        except Exception:
            setup_failure(f"Exception raised {traceback.format_exc()}", backend="aws")
            raise
    elif choice == "2":  # GCP
        print("\nRunning [green]coiled setup gcp[/green]\n")
        from .gcp import do_setup

        try:
            return do_setup()
        except Exception:
            setup_failure(f"Exception raised {traceback.format_exc()}", backend="gcp")
            raise
    elif choice == "3":  # Other
        print(
            "\nCoiled currently supports Amazon Web Service and Google Cloud."
            "It's easy to make an account with either and get "
            "started using Coiled. Please see our documentation about choosing a cloud provider:\n"
            "[link]https://docs.coiled.io/user_guide/backends.html#need-a-cloud-provider-account[/link]\n\n"
            "If you're not ready to create your own cloud provider account, "
            "but would still like to try Coiled, we can set you up with a trial on our account. "
            "Reach out to us at "
            "[link=mailto:hello@coiled.io?subject=Trial%20Account]hello@coiled.io[/link] "
            "with a brief note about how you're using Dask."
        )

    return False



from prefect import Flow, task
from prefect.executors import DaskExecutor


@task
def inc(x):
    return x + 1


@task
def double(x):
    return x * 2


@task
def add(x, y):
    return x + y


@task
def sum_all(z):
    return sum(z)


with Flow("inc-double-add-sum") as flow:
    a_range = range(10)
    a = inc.map(a_range)
    b = double.map(a_range)
    c = add.map(x=a, y=b)
    total_sum = sum_all(c)

coiled_executor = DaskExecutor(
    # tell the DaskExecutor to run on Coiled
    cluster_class="coiled.Cluster",
    # Coiled-specific keyword arguments
    cluster_kwargs={
        "n_workers": 5,
        # replace "account-name" with your account name
        "software": "prefect-example",
        # set shutdown_on_close to False to re use the cluster
        "shutdown_on_close": True,
        # name of the cluster, for easy reference
        "name": "prefect-executor",
    },
)

# run the flow
flow.run(executor=coiled_executor)

# To use with Prefect Cloud or Prefect Server:
# Register the flow under your project
# flow.register(project_name="<project-name>")



import asyncio
import sys
from abc import ABC, abstractmethod
from typing import Dict, Generic, List, Optional, TypeVar

from aiohttp import (
    ClientSession,
    ClientWebSocketResponse,
    WSMessage,
    WSMsgType,
    WSServerHandshakeError,
)

from .context import COILED_OP_CONTEXT, create_trace_data
from .errors import ServerError

MessageReturnT = TypeVar("MessageReturnT")


async def _exhaust_async_iterator(stream):
    async for _ in stream:
        pass


class WebsocketConnectorBase(Generic[MessageReturnT], ABC):
    def __init__(
        self,
        endpoint: str,
        session: ClientSession,
    ):
        self.endpoint = endpoint
        self.session = session
        self._error_messages: List[MessageReturnT] = []
        self.ws: ClientWebSocketResponse

    async def connect(self):
        self.ws = await self.session.ws_connect(self.endpoint)
        await self.ws.receive()

    async def send_json(self, data: Dict):
        try:
            await self.ws.send_json({**data, **create_trace_data()})
        except ConnectionResetError:
            # It should be okay that we just skip here, since the on_close
            # method will raise if necessary when the connection
            # is finished closing.
            print("Not sending message, connection closing")

    @abstractmethod
    async def on_message(self, message_data: Dict) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def check_error_message(self, message_data: Dict) -> Optional[MessageReturnT]:
        raise NotImplementedError()

    def error_messages(self) -> List[MessageReturnT]:
        return self._error_messages

    @abstractmethod
    async def on_error(self, message: WSMessage):
        # TODO comment about why this is different from error_messages, etc.
        raise NotImplementedError()

    @abstractmethod
    async def on_close(self, message: WSMessage):
        raise NotImplementedError()

    async def close(self, *, code: int = 1000, message: bytes = b""):
        await self.ws.close(code=code, message=message)

    def __aiter__(self):
        return self

    async def __anext__(self):
        message = await self.ws.receive()
        if message.type == WSMsgType.TEXT:
            data = message.json()
            error_message = await self.check_error_message(data)
            if error_message:
                self._error_messages.append(error_message)
            await self.on_message(data)
        elif message.type == WSMsgType.ERROR:
            await self.on_error(message)
        elif message.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
            # From what I could tell, whenever we get any one of
            # these, the ClientWebsocketResponse already has a close_code
            # and an exception (when there is one), so we should be able to
            # handle the close on any one of these message types, but I may
            # end up being wrong
            await self.on_close(message)
            raise StopAsyncIteration


class NotificationWebsocketConnector(WebsocketConnectorBase[str]):
    def __init__(
        self,
        endpoint: str,
        session: ClientSession,
        log_output=sys.stdout,
    ):
        super().__init__(endpoint, session)
        self.log_output = log_output

    async def on_message(self, message_data: Dict) -> None:
        # This could be there as a None value
        data = message_data.get("data") or {}
        # Ignore notifications that are unrelated to this context
        if not data.get("coiled-operation-id", "") == COILED_OP_CONTEXT.get(""):
            return
        message_data = message_data["message"]
        print(message_data, file=self.log_output)

    async def on_error(self, error: WSMessage) -> None:
        print(
            f"Error with connection to notifications endpoint: {error}",
            file=self.log_output,
        )

    async def check_error_message(self, message_data: Dict) -> Optional[str]:
        if message_data.get("is_error", False):
            return message_data["message"]

    async def on_close(self, message: WSMessage) -> None:
        # Right now notifications are always riding sidecar to another
        # connection. We may need to add visibility here but I didn't
        # want to clutter the user's screen with potentially confusing
        # messages. Hopefully our server logs would capture bad things
        # related to notification connections.
        pass


class WebsocketConnector(WebsocketConnectorBase[str]):
    def __init__(
        self,
        endpoint: str,
        notifications_endpoint: str,
        session: ClientSession,
        log_output=sys.stdout,
        connection_error_message: str = "Unable to connect to server",
    ):
        super().__init__(endpoint, session)
        self.log_output = log_output
        self.notification_connector = NotificationWebsocketConnector(
            endpoint=notifications_endpoint,
            session=session,
            log_output=log_output,
        )
        self.connection_error_message = connection_error_message

    async def connect(self):
        try:
            await super().connect()
            await self.notification_connector.connect()
        except WSServerHandshakeError:
            raise ValueError(self.connection_error_message)

    async def on_message(self, message_data: Dict) -> None:
        if "message" in message_data:
            message = message_data["message"]
            print(message, file=self.log_output)

    async def check_error_message(self, message_data: Dict) -> Optional[str]:
        if message_data.get("is_error") or message_data.get("raise"):
            return message_data["message"]

    async def on_error(self, message: WSMessage) -> None:
        print("Error with connection: ", str(message.extra), file=self.log_output)

    async def on_close(self, message: WSMessage) -> None:
        ws_exception = self.ws.exception()
        close_code: Optional[int] = self.ws.close_code

        if close_code == 4403:
            raise ValueError(
                "Unauthorized connection. Ensure that you are "
                "using a valid token, or that you are connecting "
                "to the correct account."
            )

        is_bad_close_code = close_code and 1000 < close_code < 3000
        error_messages = self.error_messages()

        try:
            if self.notification_connector:
                await self.notification_connector.close()
        except Exception:
            # The notification connector is not mission critical,
            pass

        if error_messages:
            # Just print the ws_exception and close code, raise the last
            # error message we sent
            if ws_exception:
                print(f"Websocket exception: {ws_exception}", file=self.log_output)
            raise ServerError(error_messages[-1])
        elif ws_exception or is_bad_close_code:
            ws_error_message = f"Error with connection to server: code {close_code}"
            if ws_exception:
                raise ValueError(ws_error_message) from ws_exception
            else:
                raise ValueError(ws_error_message)

    async def stream_messages(self):
        await asyncio.gather(
            _exhaust_async_iterator(self),
            _exhaust_async_iterator(self.notification_connector),
        )


class ConfigureBackendConnector(WebsocketConnector):
    async def on_message(self, message_data: Dict) -> None:
        if message_data.get("type") != "update_options":
            return

        message = message_data["message"]
        if message == "FINISHED":
            await self.close()
            return
        elif message in ["CONFIGURING_BACKEND", "NOT_CONFIGURING_BACKEND"]:
            # These are messages used by the web UI, we can ignore
            # them here.
            return

        print(message, file=self.log_output)



import asyncio
import json
import random
import shutil
import string
from os import environ
from typing import AsyncGenerator, NamedTuple

import dask
import pytest
import structlog
import toolz
from distributed.utils_test import loop  # noqa: F401
from django.conf import settings
from django.utils import timezone

import coiled
from api_tokens.models import ApiToken, generate_token
from backends.types import BackendChoices
from coiled.core import Async, Cloud
from coiled.utils import run_command_in_subprocess
from declarative.types import BackendTypesEnum
from pricing.models import VmTypeModel
from users.models import Account, Membership, User, get_default_registry_data

logger = structlog.get_logger(__name__)

HAS_BUILDAH = shutil.which("buildah") is not None

account_suffix = "".join(random.choices(string.ascii_lowercase + string.digits, k=8))


class UserOrgMembership(NamedTuple):
    user: User
    account: Account
    membership: Membership


PASSWORD = "mypassword"
ACCOUNT = f"test-{account_suffix}"
CONFIGURATION = "myclusterconfig"
SOFTWARE_NAME = "myenv"


class MockResponse:
    def __init__(self, json_data, status):
        self.json_data = json_data
        self.response = asyncio.Future()
        self.response.set_result({"next": None, "results": json_data})
        self.status = status

    async def json(self):
        return {"next": None, "results": self.json_data}


@pytest.fixture()
# lets keep this a fixture so we can eventually support multiple backends in a
# single run via paramters
def backend_name(request):
    yield settings.DEFAULT_CLUSTER_BACKEND


@pytest.fixture
async def cleanup(cloud):
    clusters = await cloud.list_clusters()
    await asyncio.gather(*[cloud.delete_cluster(cluster_id=cluster["id"]) for cluster in clusters])

    yield

    clusters = await cloud.list_clusters()
    await asyncio.gather(*[cloud.delete_cluster(cluster_id=cluster["id"]) for cluster in clusters])


@pytest.fixture
async def docker_prune():
    MIN_DISK_SPACE = 15.0  # GiB
    # If the disk usage has dropped below MIN_DISK_SPACE GB, prune aggressively
    used = shutil.disk_usage("/").free / 1024 ** 3
    print(f"Current disk space available: {used} GB")
    if used < MIN_DISK_SPACE:
        print(f"Free disk space has dropped to {used} GB, pruning...")
        if HAS_BUILDAH:
            async for _ in run_command_in_subprocess("buildah rmi -a -f"):
                pass
        else:
            async for _ in run_command_in_subprocess("docker rmi -f $(docker images -aq)"):
                pass
        print(f"Free disk space now at {shutil.disk_usage('/').free/1024**3} GB")
    before = set()

    # Get a list of images before the test is run.
    if HAS_BUILDAH:
        images_json = ""
        async for line in run_command_in_subprocess("buildah images --json 2>/dev/null"):
            images_json += line
        for image_data in json.loads(images_json):
            before.add(image_data["id"])
    else:
        async for l in run_command_in_subprocess("docker images --format '{{json .}}'"):
            data = json.loads(l)
            before.add(data["ID"])

    print(f"Starting images: {before}")
    yield

    # Attempt to remove new images created during the test.
    after = set()
    if HAS_BUILDAH:
        images_json = ""
        async for line in run_command_in_subprocess("buildah images --json 2>/dev/null"):
            images_json += line
        for image_data in json.loads(images_json):
            after.add(image_data["id"])
    else:
        async for l in run_command_in_subprocess("docker images --format '{{json .}}'"):
            data = json.loads(l)
            after.add(data["ID"])
    print(f"Ending images: {after}")

    new_images = after - before
    if len(new_images) == 0:
        return

    print(f"New images to be removed: {new_images}")
    cmd = "buildah" if HAS_BUILDAH else "docker"
    try:
        async for _ in run_command_in_subprocess(f"{cmd} rmi -f {' '.join(new_images)}"):
            pass
        print(f"Cleaned up images {new_images}")
    except ValueError as e:
        print(f"Failed to remove {new_images} with error {e}")

    # If the disk usage has dropped below MIN_DISK_SPACE GB, prune aggressively
    used = shutil.disk_usage("/").free / 1024 ** 3
    print(f"Current disk space available: {used} GB")
    if used < MIN_DISK_SPACE:
        print(f"Free disk space has dropped to {used} GB, pruning...")
        if HAS_BUILDAH:
            async for _ in run_command_in_subprocess("buildah rmi -a -f"):
                pass
        else:
            async for _ in run_command_in_subprocess("docker rmi -f $(docker images -aq)"):
                pass
        print(f"Free disk space now at {shutil.disk_usage('/').free/1024**3} GB")


@pytest.fixture(scope="function")
def container_registry_fixture():
    if settings.TEST_BACKEND == "in_process":
        return get_default_registry_data(BackendChoices.IN_PROCESS)
    # TODO will have to update this when we start testing multi-cloud
    return get_default_registry_data()


@pytest.fixture(scope="function")
def fedex_account(transactional_db, django_user_model, container_registry_fixture, backend_name):
    account = Account(
        slug="fedex",
        name="FedEx",
        backend=backend_name,
        options={},
        container_registry=container_registry_fixture,
        can_use_coiled_hosted=True,
    )
    account.save()
    return account


@pytest.fixture(scope="function")
def base_user_object(
    transactional_db,
    django_user_model,
    remote_access_url,
    container_registry_fixture,
    backend_name,
):
    joined = timezone.datetime(2021, 11, 7, tzinfo=timezone.get_current_timezone())
    user = django_user_model.objects.create(
        username=ACCOUNT,
        email="myuser@users.com",
        date_joined=joined,
    )
    user.set_password(PASSWORD)
    user.save()
    return user


@pytest.fixture(scope="function")
def base_user_token(base_user_object):
    secret_value, token_hash = generate_token()
    token = ApiToken.objects.create(
        user=base_user_object,
        token_hash=token_hash,
    )
    token.save()
    token = token.get_token(secret_value)
    return token


@pytest.fixture(scope="function")
def superuser(
    transactional_db,
    django_user_model,
):
    user = django_user_model.objects.create_superuser(
        username="mrsuperuser",
        email="super@user.com",
    )
    user.save()

    return user


@pytest.fixture(scope="function")
def base_user(
    transactional_db,
    remote_access_url,
    container_registry_fixture,
    backend_name,
    base_user_object,
    base_user_token,
):
    user = base_user_object
    membership = Membership.objects.filter(user=user).first()
    membership.account.container_registry = container_registry_fixture
    membership.account.backend = backend_name
    membership.account.options = {}
    membership.account.can_use_coiled_hosted = True
    membership.account.save()

    user.auth_token.delete()

    with dask.config.set(
        {
            "coiled.user": f"{user.username}",
            "coiled.token": base_user_token,
            "coiled.server": remote_access_url,
            "coiled.account": ACCOUNT,
            "coiled.backend-options": {"region": "us-east-2"},
            "coiled.no-minimum-version-check": True,
        }
    ):
        yield UserOrgMembership(user, membership.account, membership)


# fixture must be function-scoped because `transactional_db` is
@pytest.fixture(scope="function")
def sample_user(base_user, backend):
    yield base_user


@pytest.fixture(scope="function")
def disabled_free_tier_user(
    transactional_db,
    django_user_model,
    backend,
    remote_access_url,
    container_registry_fixture,
    backend_name,
):
    user = django_user_model.objects.create(
        username="kryptominer",
        email="kryptominer@users.com",
    )
    user.set_password(PASSWORD)
    user.save()
    account = Account.objects.create(
        name=f"{ACCOUNT}-kryptominer-inc",
        backend=backend_name,
        options={},
        container_registry=container_registry_fixture,
        active=False,
    )
    membership = Membership.objects.create(user=user, account=account, is_admin=True, limit=4)

    program = account.active_program
    program.core_limit = 0
    program.save()

    with dask.config.set(
        {
            "coiled.user": f"{user.username}",
            "coiled.token": f"{user.auth_token.key}",
            "coiled.server": remote_access_url,
            "coiled.account": account.slug,
            "coiled.backend-options": {"region": "us-east-2"},
            "coiled.no-minimum-version-check": True,
        }
    ):
        yield UserOrgMembership(user, membership.account, membership)


@pytest.fixture(scope="function")
def sample_user_token(base_user_token):
    """this is a duplicate of base_user_token, to match the name of sample_user
    (so that when we use it next to sample_user it's obvious to readers that this token corresponds to that user
    """
    return base_user_token


@pytest.fixture()
def long_lived_token_for_sample_user(sample_user):
    secret_value, token_hash = generate_token()
    token = ApiToken.objects.create(
        user=sample_user.user,
        token_hash=token_hash,
    )
    token.save()
    return token.get_token(secret_value)


@pytest.fixture(scope="function")
def sample_gpu_user(transactional_db, django_user_model, backend, remote_access_url):
    user = django_user_model.objects.create(
        username="mygpuuser",
        email="myuser@gpuusers.com",
    )
    user.set_password(PASSWORD)
    user.save()
    membership = Membership.objects.filter(user=user).first()
    program = membership.account.active_program
    program.gpus_limit = 4
    program.save()

    with dask.config.set(
        {
            "coiled.user": f"{user.username}",
            "coiled.token": f"{user.auth_token.key}",
            "coiled.server": remote_access_url,
            "coiled.account": "myuser",
            "coiled.backend-options": {"region": "us-east-2"},
            "coiled.no-minimum-version-check": True,
        }
    ):
        yield UserOrgMembership(user, membership.account, membership)


@pytest.fixture(scope="function")
def jess_from_fedex(transactional_db, django_user_model, backend, remote_access_url, fedex_account):
    jess = django_user_model.objects.create(
        username="jess",
        email="jess@fedex.com",
    )
    jess.set_password(PASSWORD)
    jess.save()
    fedex_membership = Membership(user=jess, account=fedex_account)
    fedex_membership.save()

    with dask.config.set(
        {
            "coiled.user": f"{jess.username}",
            "coiled.token": f"{jess.auth_token.key}",
            "coiled.server": remote_access_url,
            "coiled.account": None,
            "coiled.backend-options": {"region": "us-east-2"},
            "coiled.no-minimum-version-check": True,
        }
    ):
        yield jess


# fixture must be function-scoped because `transactional_db` is
@pytest.fixture(scope="function")
def second_user(
    transactional_db,
    django_user_model,
    backend,
    remote_access_url,
    container_registry_fixture,
    backend_name,
):
    user = django_user_model.objects.create(
        username="charlie",
        email="charlie@users.com",
    )
    user.set_password(PASSWORD)
    user.save()
    account = Account.objects.create(
        name="mycorp",
        backend=backend_name,
        options={},
        container_registry=container_registry_fixture,
    )
    membership = Membership.objects.create(user=user, account=account, is_admin=True, limit=4)
    config = dask.config.get("coiled").copy()
    config.update()
    with dask.config.set(
        {
            "coiled.user": f"{user.username}",
            "coiled.token": f"{user.auth_token.key}",
            "coiled.server": remote_access_url,
            "coiled.account": None,
            "coiled.backend-options": {"region": "us-east-2"},
            "coiled.no-minimum-version-check": True,
        }
    ):
        yield UserOrgMembership(user, membership.account, membership)


# fixture must be function-scoped because `transactional_db` is
@pytest.fixture(scope="function")
def external_aws_account_user(transactional_db, django_user_model, backend, remote_access_url):
    user = django_user_model.objects.create(
        username="externalaws",
        email="aws@users.com",
    )
    user.set_password(PASSWORD)
    user.save()

    account = Account.objects.get(slug=user.default_account.slug)
    options = {
        "credentials": {
            "aws_secret_access_key": environ.get("TEST_AWS_SECRET_ACCESS_KEY", ""),
            "aws_access_key_id": environ.get("TEST_AWS_ACCESS_KEY_ID", ""),
        },
        "account_role": environ.get("TEST_AWS_IAM_ROLE", ""),  # Optional
    }
    account.options = options
    account.save()

    with dask.config.set(
        {
            "coiled.user": f"{user.username}",
            "coiled.token": f"{user.auth_token.key}",
            "coiled.server": remote_access_url,
            "coiled.account": user.username,
            "coiled.backend-options": {"region": "us-east-2"},
            "coiled.no-minimum-version-check": True,
        }
    ):
        yield user


@pytest.fixture(scope="function")
def second_account(sample_user, container_registry_fixture, backend_name):
    account = Account.objects.create(
        name="OtherOrg",
        backend=backend_name,
        options={},
        container_registry=container_registry_fixture,
    )
    membership = Membership.objects.create(user=sample_user.user, account=account, is_admin=False, limit=2)
    sample_user.user.save()
    return UserOrgMembership(sample_user.user, account, membership)


@pytest.fixture(scope="function")
def account_with_options(sample_user, container_registry_fixture, backend_name):
    account = Account.objects.create(
        name="GotOptions",
        backend=backend_name,
        options=toolz.merge({}, {"region": "us-west-1"}),
        container_registry=container_registry_fixture,
    )
    membership = Membership.objects.create(user=sample_user.user, account=account, is_admin=False, limit=2)
    sample_user.user.save()
    return UserOrgMembership(sample_user.user, account, membership)


@pytest.fixture(scope="function")
@pytest.mark.django_db(transaction=True)  # implied by `live_server`, but explicit
async def cloud(sample_user, backend) -> AsyncGenerator[Cloud[Async], None]:
    async with coiled.Cloud(account=ACCOUNT, asynchronous=True) as cloud:
        # Remove default software environments and cluster configurations
        default_envs = await cloud.list_software_environments(new_build_backend=False)
        await asyncio.gather(
            *[
                cloud.delete_software_environment(name=name, new_build_backend=False)
                for name, info in default_envs.items()
            ]
        )
        yield cloud


@pytest.fixture(scope="function")
@pytest.mark.django_db(transaction=True)  # implied by `live_server`, but explicit
async def disabled_free_tier_cloud(disabled_free_tier_user, backend) -> AsyncGenerator[Cloud[Async], None]:
    async with coiled.Cloud(account=disabled_free_tier_user.account.slug, asynchronous=True) as cloud:
        # Remove default software environments and cluster configurations
        default_envs = await cloud.list_software_environments(new_build_backend=False)
        await asyncio.gather(
            *[
                cloud.delete_software_environment(name=name, new_build_backend=False)
                for name, info in default_envs.items()
            ]
        )
        yield cloud


@pytest.fixture(scope="function")
@pytest.mark.django_db(transaction=True)  # implied by `live_server`, but explicit
async def cloud_with_gpu(sample_gpu_user):
    async with coiled.Cloud(account=sample_gpu_user.account.slug, asynchronous=True) as cloud:
        # Remove default software environments and cluster configurations
        default_envs = await cloud.list_software_environments(new_build_backend=False)
        await asyncio.gather(
            *[
                cloud.delete_software_environment(name=name, new_build_backend=False)
                for name, info in default_envs.items()
            ]
        )
        yield cloud


@pytest.fixture
async def software_env(cloud):
    await cloud.create_software_environment(name=SOFTWARE_NAME, container="daskdev/dask:latest")

    yield f"{ACCOUNT}/{SOFTWARE_NAME}"

    await cloud.delete_software_environment(name=SOFTWARE_NAME)


@pytest.fixture
async def cloud_with_account(cleanup, jess_from_fedex, fedex_account, software_env):
    async with coiled.Cloud(
        account="fedex",
        token=jess_from_fedex.auth_token.key,
        user="jess",
        asynchronous=True,
    ) as cloud:
        yield cloud


@pytest.fixture(scope="function")
def account_with_env_variables(sample_user, container_registry_fixture, backend_name):
    account = Account.objects.create(
        name=f"GotEnvs-{account_suffix}",
        backend=backend_name,
        options={},
        environment_variables={"MY_TESTING_ENV": "env_variable"},
        container_registry=container_registry_fixture,
    )
    membership = Membership.objects.create(user=sample_user.user, account=account, is_admin=False, limit=20)
    sample_user.user.save()

    return UserOrgMembership(sample_user.user, account, membership)


@pytest.fixture
def api_instance_types_gcp():
    return {
        "t2d-standard-8": {
            "name": "t2d-standard-8",
            "cores": 8,
            "gpus": 0,
            "gpu_name": None,
            "memory": 32768,
            "backend_type": "vm_gcp",
        },
        "t2d-standard-60": {
            "name": "t2d-standard-60",
            "cores": 60,
            "gpus": 0,
            "gpu_name": None,
            "memory": 245760,
            "backend_type": "vm_gcp",
        },
        "t2d-standard-48": {
            "name": "t2d-standard-48",
            "cores": 48,
            "gpus": 0,
            "gpu_name": None,
            "memory": 196608,
            "backend_type": "vm_gcp",
        },
        "t2d-standard-4": {
            "name": "t2d-standard-4",
            "cores": 4,
            "gpus": 0,
            "gpu_name": None,
            "memory": 16384,
            "backend_type": "vm_gcp",
        },
        "t2d-standard-32": {
            "name": "t2d-standard-32",
            "cores": 32,
            "gpus": 0,
            "gpu_name": None,
            "memory": 131072,
            "backend_type": "vm_gcp",
        },
        "t2d-standard-2": {
            "name": "t2d-standard-2",
            "cores": 2,
            "gpus": 0,
            "gpu_name": None,
            "memory": 8192,
            "backend_type": "vm_gcp",
        },
        "t2d-standard-16": {
            "name": "t2d-standard-16",
            "cores": 16,
            "gpus": 0,
            "gpu_name": None,
            "memory": 65536,
            "backend_type": "vm_gcp",
        },
        "g1-small": {
            "name": "g1-small",
            "cores": 1,
            "gpus": 0,
            "gpu_name": None,
            "memory": 1740,
            "backend_type": "vm_gcp",
        },
        "f1-micro": {
            "name": "f1-micro",
            "cores": 1,
            "gpus": 0,
            "gpu_name": None,
            "memory": 614,
            "backend_type": "vm_gcp",
        },
    }


@pytest.fixture(autouse=True, scope="function")
@pytest.mark.asyncio
async def set_aws_role(set_aws_role):
    # This is just here to autouse the fixture from parent conftest.
    return set_aws_role


@pytest.fixture(scope="function")
def event_loop():
    loop = asyncio.new_event_loop()  # noqa: F811
    asyncio.set_event_loop(asyncio.new_event_loop())
    yield loop
    loop.close()


@pytest.fixture(autouse=True)
def api_vm_types_aws(db):
    vm, _ = VmTypeModel.objects.get_or_create(
        name="t3.medium",
        backend_type=BackendTypesEnum.VM_AWS,
        cores=2,
        memory=4096,
    )
    vm1, _ = VmTypeModel.objects.get_or_create(
        name="t3.small",
        backend_type=BackendTypesEnum.VM_AWS,
        cores=2,
        memory=2048,
    )
    vm2, _ = VmTypeModel.objects.get_or_create(
        name="t3.large",
        backend_type=BackendTypesEnum.VM_AWS,
        cores=2,
        memory=8192,
    )
    vm3, _ = VmTypeModel.objects.get_or_create(
        name="t3a.small",
        backend_type=BackendTypesEnum.VM_AWS,
        cores=2,
        memory=2048,
    )
    vm4, _ = VmTypeModel.objects.get_or_create(
        name="t3a.medium",
        backend_type=BackendTypesEnum.VM_AWS,
        cores=2,
        memory=4096,
    )
    return [vm, vm1, vm2, vm3, vm4]



import pytest
from distributed.utils_test import loop  # noqa: F401

import coiled


@pytest.fixture
async def three_clouds(cleanup, sample_user, jess_from_fedex, remote_access_url, settings, base_user_token):
    """
    Return a tuple of three clouds, one for coiled, one for jess,
    and one for fedex. Each will have one environment and one configuration
    """
    # coiled is not a member of FedEx
    coiled_user = sample_user.user
    jess = jess_from_fedex

    async with coiled.Cloud(
        user=coiled_user.username,
        token=base_user_token,
        asynchronous=True,
    ) as my_cloud:
        async with coiled.Cloud(
            user=jess.username,
            token=jess.auth_token.key,
            asynchronous=True,
        ) as jesss_cloud:
            async with coiled.Cloud(
                user=jess.username,
                token=jess.auth_token.key,
                account="fedex",
                asynchronous=True,
            ) as fedex_cloud:
                await my_cloud._create_software_environment(
                    name="env1",
                    container="daskdev/dask:latest",
                    new_build_backend=False,
                )
                await jesss_cloud._create_software_environment(
                    name="env1",
                    container="daskdev/dask:latest",
                    new_build_backend=False,
                )
                await fedex_cloud._create_software_environment(
                    name="fedex/env1",
                    container="daskdev/dask:latest",
                    new_build_backend=False,
                )
                yield my_cloud, jesss_cloud, fedex_cloud


@pytest.mark.asyncio
async def test_can_create_software_environment_using_full_name_or_short_name(
    three_clouds,
):
    _, _, fedex = three_clouds

    # First with fully qualified name
    await fedex.create_software_environment(name="goodstuff", container="daskdev/dask:latest", new_build_backend=False)
    configs = await fedex.list_software_environments(new_build_backend=False)
    e = configs["fedex/goodstuff"]
    assert e["account"] == "fedex"

    # Then with short name
    await fedex.create_software_environment(name="goodstuff2", container="daskdev/dask:latest", new_build_backend=False)
    configs = await fedex.list_software_environments(new_build_backend=False)
    e = configs["fedex/goodstuff2"]
    assert e["account"] == "fedex"


@pytest.mark.asyncio
async def test_can_create_software_environment_in_different_account_when_member(
    three_clouds,
):
    _, jesss_cloud, fedex = three_clouds
    # jess is a member of fedex, so she should be allowed to create in that account
    await jesss_cloud.create_software_environment(
        name="fedex/goodenv", container="daskdev/dask:latest", new_build_backend=False
    )
    jess_envs = await jesss_cloud.list_software_environments(new_build_backend=False)
    fedex_envs = await fedex.list_software_environments(new_build_backend=False)
    assert "fedex/goodenv" not in jess_envs
    assert "fedex/goodenv" in fedex_envs
    e = fedex_envs["fedex/goodenv"]
    assert e["account"] == "fedex"


@pytest.mark.asyncio
async def test_cannot_create_software_environment_in_foreign_account(three_clouds):
    # "coiled" is not a member of fedex or jesss
    my_cloud, jesss_cloud, _ = three_clouds

    with pytest.raises(ValueError, match="Unauthorized"):
        await my_cloud.create_software_environment(
            name="fedex/invasive_env",
            container="daskdev/dask:latest",
            new_build_backend=False,
        )



from __future__ import annotations

from types import ModuleType
from typing import Optional

import dask
import urllib3
from dask.distributed import Client, get_client
from dask.utils import has_keyword

import coiled


def register(client: Optional[Client] = None):
    """
    Register Coiled analytics with your Dask cluster.

    Parameters
    ----------
    client:
        The Dask client that you want to connect to
        This will use the most recently created Dask client by default
        (usually what you want anyway)

    Examples
    --------
    >>> from dask.distributed import Client
    >>> client = Client()

    >>> import coiled.analytics
    >>> coiled.analytics.register()

    Or use this module as a preload

    $ dask-scheduler --preload coiled.analytics
    """
    with coiled.Cloud():  # verify that we can log in
        pass
    client = client or get_client()

    url = dask.config.get("coiled.server") + "/api/v2/analytics/preload"
    with urllib3.PoolManager() as http:
        response = http.request("GET", url)

    if response.status != 200:
        raise Exception("Unable to collect Coiled plugin", response.status)

    compiled = compile(response.data, url, "exec")
    module = ModuleType(url)
    exec(compiled, module.__dict__)
    plugin = module.CoiledTelemetry()  # type: ignore

    if has_keyword(client.register_scheduler_plugin, "idempotent"):  # type: ignore
        client.register_scheduler_plugin(plugin, idempotent=True)  # type: ignore
    else:
        client.register_scheduler_plugin(plugin)  # type: ignore


def list_clusters(
    account: Optional[str] = None,
    since: Optional[str] = "7 days",
    user: Optional[str] = None,
):
    """List clusters associated to this account

    Clusters arrive in reverse chronological order (the first element is the
    most recent).

    Parameters
    ----------
    account:
        The account whose clusters you want to list
        You must be a member of this account
        Your default account will be used if none is provided
    since:
        The amount of time to go back in history to list clusters.
        Defaults to seven days.  Leave as ``None`` to collect all history.
        Accepts any value parseable by dask.utils.parse_timedelta
    user:
        Optionally filter on username
        Providing ``None`` selects all users

    Examples
    --------
    >>> import coiled.analytics
    >>> coiled.analytics.list_clusters()  # doctest: +SKIP
    >>> coiled.analytics.list_clusters(since="30 days", user="alice")  # doctest: +SKIP
    """
    with coiled.Cloud() as c:
        return c.list_dask_scheduler(account=account, since=since, user=user)


def list_computations(cluster_id: int, account: Optional[str] = None):
    """List computations associated to a cluster

    Parameters
    ----------
    cluster_id:
        The identifier of the cluster that you want to select
    account:
        The account whose clusters you want to list
        You must be a member of this account
        Your default account will be used if none is provided

    Examples
    --------
    >>> import coiled.analytics
    >>> clusters = coiled.analytics.list_clusters()  # doctest: +SKIP
    >>> coiled.analytics.list_computations(clusters[0]["id"])  # doctest: +SKIP

    See Also
    --------
    list_clusters
    """
    with coiled.Cloud() as c:
        return c.list_computations(cluster_id, account=account)


def list_events(cluster_id: int, account: Optional[str] = None):
    """List events associated to a cluster

    Parameters
    ----------
    cluster_id:
        The identifier of the cluster that you want to select
    account:
        The account whose clusters you want to list
        You must be a member of this account
        Your default account will be used if none is provided

    Examples
    --------
    >>> import coiled.analytics
    >>> clusters = coiled.analytics.list_clusters()  # doctest: +SKIP
    >>> coiled.analytics.list_events(clusters[0]["id"])  # doctest: +SKIP

    See Also
    --------
    list_clusters
    """
    with coiled.Cloud() as c:
        return c.list_events(cluster_id, account=account)


def close(cluster_id: int, account: Optional[str] = None):
    """Close a cluster

    This sends a request to the remote scheduler asking it to shut down.
    It calls Scheduler.close on that machine.
    It is common (although not guaranteed) that workers should shut themselves
    down after a suitable time.

    Parameters
    ----------
    cluster_id:
        The identifier of the cluster that you want to select
    account:
        The account whose clusters you want to list
        You must be a member of this account
        Your default account will be used if none is provided

    Examples
    --------
    >>> import coiled.analytics
    >>> clusters = coiled.analytics.list_clusters()  # doctest: +SKIP
    >>> coiled.analytics.close(clusters[0]["id"])  # doctest: +SKIP

    See Also
    --------
    list_clusters
    """
    with coiled.Cloud() as c:
        return c.send_state(cluster_id, account=account, desired_status="stopped")


def list_exceptions(
    account: Optional[str] = None,
    since: Optional[str] = "7 days",
    user: Optional[str] = None,
    cluster_id: Optional[int] = None,
    scheduler_id: Optional[int] = None,
):
    """List user exceptions

    Parameters
    ----------
    account (optional):
        The account whose clusters you want to list
        You must be a member of this account
        Your default account will be used if none is provided
    since (optional):
        The amount of time to go back in history to list clusters.
        Defaults to seven days.  Leave as ``None`` to collect all history.
        Accepts any value parseable by dask.utils.parse_timedelta
    user (optional):
        Optionally filter on username
        Providing ``None`` selects all users
    cluster_id (optional):
        The identifier of the cluster that you want to select
    scheduler_id (optional):
        The identifier of the scheduler analytics that you want to select

    Examples
    --------
    >>> import coiled.analytics
    >>> coiled.analytics.list_exceptions()  # doctest: +SKIP
    >>> coiled.analytics.list_exceptions(since="30 days", user="alice")  # doctest: +SKIP
    """
    with coiled.Cloud() as c:
        return c.list_exceptions(
            account=account,
            since=since,
            user=user,
            cluster_id=cluster_id,
            scheduler_id=scheduler_id,
        )


async def dask_setup(scheduler):
    url = dask.config.get("coiled.server") + "/api/v2/analytics/preload"
    with urllib3.PoolManager() as http:
        response = http.request("GET", url)

    if response.status != 200:
        raise Exception("Unable to collect Coiled plugin", response.status)

    compiled = compile(response.data, url, "exec")
    module = ModuleType(url)
    exec(compiled, module.__dict__)
    plugin = module.CoiledTelemetry()  # type: ignore
    if plugin.name in scheduler.plugins:
        return

    scheduler.add_plugin(plugin, name=plugin.name, idempotent=True)
    await plugin.start(scheduler)



from click.testing import CliRunner

from coiled.cli import diagnostics


def test_diagnostics_command(sample_user):
    runner = CliRunner()
    result = runner.invoke(diagnostics.diagnostics)
    if result.exception:
        raise result.exception
    assert result.exit_code == 0
    assert "Performing health check" in result.output
    assert "health_check" in result.output
    assert "Getting user information" in result.output
    assert "user_information" in result.output



from __future__ import annotations  # this means we get FutureRefs

from typing import List, Optional, TypedDict

from ..utils import validate_type


class FirewallOptions(TypedDict):
    # if there's firewall options, then ports and cidr are both required
    ports: List[int]
    cidr: str


class BackendOptions(TypedDict, total=False):
    region_name: Optional[str]
    zone_name: Optional[str]
    firewall: Optional[FirewallOptions]
    ingress: Optional[List[FirewallOptions]]
    spot: Optional[bool]
    spot_on_demand_fallback: Optional[bool]
    multizone: Optional[bool]
    use_dashboard_public_ip: Optional[bool]
    prometheus_write: Optional[dict]


class AWSOptions(BackendOptions, total=False):
    keypair_name: Optional[str]
    use_placement_group: Optional[bool]
    # future: add stuff like "spot" when supported


class GCPOptions(BackendOptions, total=False):
    worker_accelerator_count: Optional[int]
    worker_accelerator_type: Optional[str]


BackendOptionTypes = [AWSOptions, GCPOptions]


def test_aws_options_validation():
    assert validate_type(AWSOptions, AWSOptions())
    assert validate_type(AWSOptions, AWSOptions(region_name="foo", keypair_name="bar"))


def test_aws_nested_options_validation():
    assert validate_type(
        AWSOptions,
        AWSOptions(region_name="foo", firewall=FirewallOptions(ports=[123], cidr="0.0.0.0/0")),
    )


def test_aws_dict_validation():
    assert validate_type(AWSOptions, dict())
    assert validate_type(AWSOptions, dict(region_name="foo"))


def test_aws_nested_dict_validation():
    assert validate_type(
        AWSOptions,
        dict(region_name="foo", firewall=dict(ports=[123], cidr="0.0.0.0/0")),
    )


def test_aws_invalid_options_validation():
    assert not validate_type(AWSOptions, dict(foo="foo"))
    assert not validate_type(AWSOptions, dict(region_name="foo", foo="foo"))


def test_gcp_options_validation():
    assert validate_type(GCPOptions, GCPOptions())
    assert validate_type(GCPOptions, GCPOptions(region_name="foo"))


def test_gcp_nested_options_validation():
    assert validate_type(
        GCPOptions,
        GCPOptions(region_name="foo", firewall=FirewallOptions(ports=[123], cidr="0.0.0.0/0")),
    )

    assert not validate_type(GCPOptions, AWSOptions(region_name="foo", keypair_name="bar"))


def test_any_option_validation():
    opts = dict(region_name="foo", firewall=dict(ports=[123], cidr="0.0.0.0/0"))
    assert any((validate_type(t, opts) for t in BackendOptionTypes))

    opts = {
        "region_name": "us-east1",
        "zone_name": "us-east1-d",
        "firewall": {"ports": [8787, 8786, 22], "cidr": "0.0.0.0/0"},
    }
    assert any((validate_type(t, opts) for t in BackendOptionTypes))


def test_invalid_firewall_options_validation():
    assert not validate_type(AWSOptions, dict(region_name="foo", firewall=dict(cidr="0.0.0.0/0")))
